"""
The database access layer for hippo. Uses MongoDB and Beanie.
"""

from datetime import datetime
from enum import Enum

import pymongo
from beanie import BackLink, Document, Indexed, Link, PydanticObjectId
from pydantic import BaseModel, Field

from hippometa import ALL_METADATA_TYPE


class CollectionPolicy(Enum):
    # What to do when versions are revved of products.
    # Keep track of all versions of the product in the collection.
    ALL = "all"
    # Keep track of all new versions of the product in the collection.
    # E.g. if v2 is added to a collection, v1 is _not_ but all future
    # versions will be tracked as part of that collection.
    NEW = "new"
    # Keep track of only the 'current' version of the product.
    CURRENT = "current"
    # Keep track of only a 'fixed' version of the product. So if v2 is
    # added to the collection, and v3 is created, only v2 is tracked in
    # the collection.
    FIXED = "fixed"


class User(Document):
    name: Indexed(str, unique=True)
    email: str | None
    last_access_time: datetime | None


class FileMetadata(BaseModel):
    """
    Object containing the metadata from a single file.
    """

    id: PydanticObjectId
    name: str
    description: str | None = None
    uploader: str
    uuid: str
    bucket: str
    size: int
    checksum: str
    slug: str = Field(default="data")
    available: bool = True


class File(Document, FileMetadata):
    # Information for multi-part uploads (private)
    multipart: bool = False
    number_of_parts: int = 1
    upload_id: str | None = None
    multipart_batch_size: int | None = None
    multipart_closed: bool = False

    def to_metadata(self) -> FileMetadata:
        return FileMetadata(
            id=self.id,
            name=self.name,
            description=self.description,
            uploader=self.uploader,
            uuid=self.uuid,
            bucket=self.bucket,
            size=self.size,
            slug=self.slug,
            checksum=self.checksum,
            available=self.available,
        )


class ProductMetadata(BaseModel):
    """
    Object containing the metadata from a single version of a product.
    """

    id: PydanticObjectId

    name: str
    description: str
    metadata: ALL_METADATA_TYPE

    uploaded: datetime
    updated: datetime

    current: bool
    version: str

    sources: dict[str, FileMetadata]  # Sources are indexed by their slug
    owner: str

    replaces: str | None

    readers: list[str] = []
    writers: list[str] = []

    child_of: list[PydanticObjectId]
    parent_of: list[PydanticObjectId]

    collections: list[PydanticObjectId]

    pinned: bool = False


class ProtectedDocument(Document):
    readers: list[str] = Field(default_factory=list)
    writers: list[str] = Field(default_factory=lambda: ["admin"])
    owner: str

    current: bool = True
    replaces: Link["ProtectedDocument"] | None = None


class Product(ProtectedDocument, ProductMetadata):
    name: Indexed(str, pymongo.TEXT)

    sources: dict[str, File]

    replaces: Link["Product"] | None = None

    child_of: list[Link["Product"]] = []
    parent_of: list[BackLink["Product"]] = Field(
        json_schema_extra={"original_field": "child_of"}, default=[]
    )

    collections: list[Link["Collection"]] = []
    collection_policies: list[CollectionPolicy] = [CollectionPolicy.CURRENT]

    async def to_metadata(self) -> ProductMetadata:
        # If links are longer than the pre-determined limit (3) we need
        # to fetch.
        if self.replaces is None:
            replaces_version = None
        elif isinstance(self.replaces, Link):
            replaces_version = (await self.replaces.fetch(fetch_links=True)).version
        else:
            replaces_version = self.replaces.version

        return ProductMetadata(
            id=self.id,
            name=self.name,
            description=self.description,
            metadata=self.metadata,
            uploaded=self.uploaded,
            updated=self.updated,
            current=self.current,
            version=self.version,
            sources={x: y.to_metadata() for x, y in self.sources.items()},
            owner=self.owner,
            readers=self.readers,
            writers=self.writers,
            replaces=replaces_version,
            child_of=[x.id for x in self.child_of],
            # Filter out backlinks as they are unfetchable; this only occurs
            # when walking the product tree.
            parent_of=[x.id for x in self.parent_of if not isinstance(x, BackLink)],
            collections=[x.id for x in self.collections],
        )


class CollectionMetadata(BaseModel):
    """
    Base model for a collection.
    """

    id: PydanticObjectId

    name: str
    description: str

    products: list[ProductMetadata]
    child_collections: list[PydanticObjectId]
    parent_collections: list[PydanticObjectId]

    pinned: bool = False


class Collection(ProtectedDocument, CollectionMetadata):
    # TODO: Implement updated time for collections.

    name: Indexed(str, pymongo.TEXT)
    products: list[BackLink[Product]] = Field(
        json_schema_extra={"original_field": "collections"}
    )
    child_collections: list[Link["Collection"]] = []
    parent_collections: list[BackLink["Collection"]] = Field(
        json_schema_extra={"original_field": "child_collections"}
    )


BEANIE_MODELS = [User, File, Product, Collection]



"""
Models for relationships and collections.
"""

import datetime

from beanie import PydanticObjectId
from pydantic import BaseModel

from hippometa import ALL_METADATA_TYPE


class CreateCollectionRequest(BaseModel):
    """
    Request model for creating a collection.
    """

    description: str
    readers: list[str] = []
    writers: list[str] = []


class UpdateCollectionRequest(BaseModel):
    """
    Request model for updating a collection.
    """

    name: str | None = None
    description: str | None = None
    owner: str | None = None
    add_readers: list[str] = []
    remove_readers: list[str] = []
    add_writers: list[str] = []
    remove_writers: list[str] = []


class ReadCollectionProductResponse(BaseModel):
    """
    Response model for reading a product in a collection.
    """

    id: PydanticObjectId
    name: str
    description: str
    owner: str
    version: str
    uploaded: datetime.datetime
    metadata: ALL_METADATA_TYPE


class ReadCollectionCollectionResponse(BaseModel):
    """
    Response model for reading a collection (i.e. a parent or child)
    as part of a collection request.
    """

    id: PydanticObjectId
    name: str
    description: str
    owner: str
    readers: list[str]
    writers: list[str]


class ReadCollectionResponse(BaseModel):
    """
    Response model for reading a collection.
    """

    id: PydanticObjectId
    name: str
    description: str
    products: list[ReadCollectionProductResponse] | None
    child_collections: list[ReadCollectionCollectionResponse] | None = None
    parent_collections: list[ReadCollectionCollectionResponse] | None = None
    readers: list[str]
    writers: list[str]
    owner: str



from typing import Literal

from hippometa.base import BaseMetadata


class CatalogMetadata(BaseMetadata):
    """
    A catalog (potentially of resolved sources).
    """

    metadata_type: Literal["catalog"] = "catalog"

    file_type: Literal["csv", "fits", "hdf5", "txt"]
    column_description: dict[str, str] | None = None

    telescope: str | None = None
    instrument: str | None = None
    release: str | None = None



"""
A beam.
"""

from typing import Literal

from hippometa.base import BaseMetadata


class BeamMetadata(BaseMetadata):
    """
    Metadata for a beam object.
    """

    metadata_type: Literal["beam"] = "beam"

    telescope: str | None = None
    "Telescope that this beam is for"
    instrument: str | None = None
    "Instrument this beam is for (e.g. optics tube)"
    wafer: str | None = None
    "Wafer that this beam is for"
    frequency: str | None = None
    "Frequency that this beam is for"



# from abc import ABC, abstractmethod
from pathlib import Path

from pydantic import BaseModel, Field

from hippoclient.caching import MultiCache


class SourceInstance(BaseModel):
    path: Path
    slug: str | None = None
    name: str | None = None
    description: str | None = None


class LocalSource(SourceInstance):
    def model_post_init(self, __context):
        if self.name is None:
            self.name = self.path.stem

        return super().model_post_init(__context)

    def __repr__(self):
        return (
            f"LocalSource(path={self.path.__repr__()}, slug='{self.slug}',"
            f"name='{self.name}', description='{self.description}')"
        )

    def __str__(self):
        return f"LocalSource {self.slug} ({self.name}; {self.description}) representing {self.path}"


class RemoteSource(SourceInstance):
    path: Path | None = None

    source_id: str = Field(frozen=True)

    cached: bool = False
    cache: MultiCache | None = None

    realize: bool = False

    def model_post_init(self, __context):
        if self.realize:
            self.path = self.__realize()
        return super().model_post_init(__context)

    def __realize(self) -> Path:
        if self.path is not None:
            return self.path

        if self.cache is not None:
            if path := self.cache.available(self.source_id):
                self.path = path
                self.cached = True
                return self.path
            else:
                raise RuntimeError(
                    "Attempted to realize a source that was not cached; this is not yet supported"
                )
        else:
            raise AttributeError(
                "RemoteSource has no cache set; cannot realize without a cache"
            )

    def __repr__(self):
        return (
            f"RemoteSource(path={self.path.__repr__()}, slug='{self.slug}',"
            f"name='{self.name}', description='{self.description}', cached={self.cached}, "
            f"cache={self.cache.__repr__()}, source_id='{self.source_id}')"
        )

    def __str__(self):
        if self.cached:
            return f"Realized RemoteSource {self.slug} representing {self.path}"
        else:
            return f"Unrealized RemoteSource {self.slug}: {self.name}"



"""
Metadata for site camera timelapse videos.
"""

from datetime import datetime
from typing import ClassVar, Literal

from .base import BaseMetadata


class CameraMetadata(BaseMetadata):
    """
    Metadata for site camera timelapse videos.
    """

    metadata_type: Literal["camera"] = "camera"

    date: datetime
    "The date of the camera videos"

    valid_slugs: ClassVar[set[str]] = {
        "act_highbay",
        "busbar",
        "c1",
        "c2",
        "c2_front",
        "c3",
        "c3_front",
        "c4",
        "c4_front",
        "c5",
        "highbay_back",
        "highbay_cargo",
        "lat_g5",
        "lat_highbay",
        "lat_shutter",
        "latr",
        "mirror1_latr",
        "mirror2_latr",
        "pumphouse",
        "pumphouse_inside",
        "satp1",
        "satp2",
        "satp3",
        "site_entrance",
    }



"""
API endpoints for relationships between products and collections.
"""

from difflib import Differ

from beanie import PydanticObjectId
from fastapi import APIRouter, HTTPException, Request, status
from loguru import logger

from hipposerve.api.models.relationships import (
    CreateCollectionRequest,
    ReadCollectionCollectionResponse,
    ReadCollectionProductResponse,
    ReadCollectionResponse,
    UpdateCollectionRequest,
)
from hipposerve.service import acl, collection, product
from hipposerve.service.auth import requires

relationship_router = APIRouter(prefix="/relationships")


@relationship_router.put("/collection/{name}")
@requires(["hippo:admin", "hippo:write"])
async def create_collection(
    name: str,
    model: CreateCollectionRequest,
    request: Request,
) -> PydanticObjectId:
    """
    Create a new collection with {name}.
    """

    logger.info(
        "Request to create collection: {} from {}", name, request.user.display_name
    )

    coll = await collection.create(
        name=name,
        user=request.user.display_name,
        description=model.description,
        collection_readers=model.readers,
        collection_writers=model.writers,
    )

    logger.info(
        "Collection {} ({}) created for {}", coll.id, name, request.user.display_name
    )

    return coll.id


@relationship_router.get("/collection/{id}")
@requires(["hippo:admin", "hippo:read"])
async def read_collection(
    id: PydanticObjectId,
    request: Request,
) -> ReadCollectionResponse:
    """
    Read a collection's details.
    """

    logger.info("Request to read collection: {} from {}", id, request.user.display_name)

    try:
        item = await collection.read(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )

    return ReadCollectionResponse(
        id=item.id,
        name=item.name,
        description=item.description,
        owner=item.owner,
        readers=item.readers,
        writers=item.writers,
        products=[
            ReadCollectionProductResponse(
                id=x.id,
                name=x.name,
                version=x.version,
                description=x.description,
                owner=x.owner,
                uploaded=x.uploaded,
                metadata=x.metadata,
            )
            for x in item.products
        ],
        child_collections=[
            ReadCollectionCollectionResponse(
                id=x.id,
                name=x.name,
                description=x.description,
                owner=x.owner,
                readers=x.readers,
                writers=x.writers,
            )
            for x in item.child_collections
        ],
        parent_collections=[
            ReadCollectionCollectionResponse(
                id=x.id,
                name=x.name,
                description=x.description,
                owner=x.owner,
                readers=x.readers,
                writers=x.writers,
            )
            for x in item.parent_collections
        ],
    )


@relationship_router.get("/collection/search/{name}")
@requires(["hippo:admin", "hippo:read"])
async def search_collection(
    name: str,
    request: Request,
) -> list[ReadCollectionResponse]:
    """
    Search for collections by name. Products and sub-collections are not
    returned; these should be fetched separately through the read_collection
    endpoint.
    """

    logger.info(
        "Request to search for collection: {} from {}", name, request.user.display_name
    )

    results = await collection.search_by_name(
        name=name, groups=request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Found {} collections for {} from {}",
        len(results),
        name,
        request.user.display_name,
    )

    return [
        ReadCollectionResponse(
            id=item.id,
            name=item.name,
            description=item.description,
            owner=item.owner,
            readers=item.readers,
            writers=item.writers,
            products=None,
        )
        for item in results
    ]


@relationship_router.put("/collection/{collection_id}/{product_id}")
@requires(["hippo:admin", "hippo:write"])
async def add_product_to_collection(
    collection_id: PydanticObjectId,
    product_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Add a product to a collection.
    """

    logger.info(
        "Request to add product {} to collection {} from {}",
        product_id,
        collection_id,
        request.user.display_name,
    )

    try:
        coll = await collection.read(
            id=collection_id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )

    try:
        item = await product.read_by_id(
            id=product_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        await product.add_collection(
            product=item,
            access_groups=request.user.groups,
            collection=coll,
            scopes=request.auth.scopes,
        )
        logger.info("Successfully added {} to collection {}", item.name, coll.name)
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )


@relationship_router.delete("/collection/{collection_id}/{product_id}")
@requires(["hippo:admin", "hippo:write"])
async def remove_product_from_collection(
    collection_id: PydanticObjectId,
    product_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Remove a product from a collection.
    """

    logger.info(
        "Request to remove product {} from collection {} from {}",
        product_id,
        collection_id,
        request.user.display_name,
    )

    try:
        coll = await collection.read(
            id=collection_id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )

    try:
        item = await product.read_by_id(
            id=product_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        await product.remove_collection(
            product=item,
            access_groups=request.user.groups,
            collection=coll,
            scopes=request.auth.scopes,
        )
        logger.info("Successfully removed {} from collection {}", item.name, coll.name)
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )


@relationship_router.post("/collection/{id}/diff")
@requires(["hippo:admin", "hippo:write"])
async def diff_collection(
    id: PydanticObjectId,
    model: UpdateCollectionRequest,
    request: Request,
) -> dict[str, list[str] | str]:
    """
    Diff a collection's current state with the proposed changes.
    """

    logger.info("Request to diff collection: {} from {}", id, request.user.display_name)

    diff = await collection.diff(
        id=id,
        access_groups=request.user.groups,
        scopes=request.auth.scopes,
        name=model.name,
        description=model.description,
    )

    if diff.get("description", None) is not None:
        differ = Differ()
        diffed_description = list(
            differ.compare(
                diff["description"][0].splitlines(),
                diff["description"][1].splitlines(),
            )
        )

        output_diff = "<table class='table table-bordered'><tbody>\n"
        for line, next_line in zip(diffed_description, diffed_description[1:] + [""]):
            if line.startswith("? "):
                continue
            if line.startswith("+ "):
                line = line[2:]
                if next_line.startswith("? "):
                    # Two lines: first is addition, second is location in the original line
                    position = next_line[2:]

                    new_line = ""
                    in_diff = False

                    for outchar, indicator in zip(line, position + " " * len(line)):
                        if indicator != " ":
                            if not in_diff:
                                new_line += "<b class='text-success'>"
                                in_diff = True
                                new_line += outchar
                                continue
                            else:
                                new_line += outchar
                                continue

                        if in_diff:
                            new_line += "</b>"
                            in_diff = False

                        new_line += outchar

                    line = new_line
                output_diff += f"<tr class='table-success'><td><b class='text-success'>+</b></td><td class='text-success'>{line}</td></tr>\n"
            elif line.startswith("- "):
                line = line[2:]
                if next_line.startswith("? "):
                    # Two lines: first is addition, second is location in the original line
                    position = next_line[2:]

                    new_line = ""
                    in_diff = False

                    for outchar, indicator in zip(line, position + " " * len(line)):
                        if indicator != " ":
                            if not in_diff:
                                new_line += "<b class='text-danger'>"
                                in_diff = True
                                new_line += outchar
                                continue
                            else:
                                new_line += outchar
                                continue

                        if in_diff:
                            new_line += "</b>"
                            in_diff = False

                        new_line += outchar

                    line = new_line
                output_diff += f"<tr class='table-danger'><td><b class='text-danger'>-</b></td><td class='text-danger'>{line}</td></tr>\n"
            else:
                output_diff += f"<tr><td></td><td>{line[2:]}</td></tr>\n"

        output_diff += "</tbody></table>\n"
        diff["description"] = output_diff

    logger.info(
        "Successfully diffed collection {} from {}", id, request.user.display_name
    )
    return diff


@relationship_router.post("/collection/{id}")
@requires(["hippo:admin", "hippo:write"])
async def update_collection(
    id: PydanticObjectId,
    model: UpdateCollectionRequest,
    request: Request,
) -> PydanticObjectId:
    logger.info(
        "Request to update collection: {} from {}", id, request.user.display_name
    )
    coll = await collection.update(
        id=id,
        access_groups=request.user.groups,
        scopes=request.auth.scopes,
        name=model.name,
        description=model.description,
    )

    coll = await acl.update_access_control(
        doc=coll,
        owner=model.owner,
        add_readers=model.add_readers,
        remove_readers=model.remove_readers,
        add_writers=model.add_writers,
        remove_writers=model.remove_writers,
    )

    logger.info(
        "Successfully updated collection {} from {}", id, request.user.display_name
    )
    return coll.id


@relationship_router.delete("/collection/{id}")
@requires(["hippo:admin", "hippo:write"])
async def delete_collection(
    id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Delete a collection.
    """

    logger.info(
        "Request to delete collection: {} from {}", id, request.user.display_name
    )

    try:
        # Check if we have a parent; if we do, we need to remove its link to us.
        coll = await collection.read(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )

        if coll.parent_collections:
            for parent in coll.parent_collections:
                await collection.remove_child(
                    parent_id=parent.id,
                    child_id=id,
                    groups=request.user.groups,
                    scopes=request.auth.scopes,
                )

        await collection.delete(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
        logger.info(
            "Successfully deleted collection {} from {}", id, request.user.display_name
        )
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )


@relationship_router.put("/product/{parent_id}/child_of/{child_id}")
@requires(["hippo:admin", "hippo:write"])
async def add_child_product(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Add a child product to a parent product.
    """

    logger.info(
        "Request to add product {} as child of {} from {}",
        child_id,
        parent_id,
        request.user.display_name,
    )

    try:
        source = await product.read_by_id(
            id=parent_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        destination = await product.read_by_id(
            id=child_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        await product.add_relationship(
            source=source,
            destination=destination,
            access_groups=request.user.groups,
            type="child",
            scopes=request.auth.scopes,
        )
        logger.info(
            "Successfully added {} as child of {}", destination.name, source.name
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )


@relationship_router.delete("/product/{parent_id}/child_of/{child_id}")
@requires(["hippo:admin", "hippo:write"])
async def remove_child_product(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Remove a parent-child relationship between two products.
    """

    logger.info(
        "Request to remove product {} as child of {} from {}",
        child_id,
        parent_id,
        request.user.display_name,
    )

    try:
        source = await product.read_by_id(
            id=parent_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        destination = await product.read_by_id(
            id=child_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        await product.remove_relationship(
            source=source,
            destination=destination,
            access_groups=request.user.groups,
            type="child",
            scopes=request.auth.scopes,
        )
        logger.info(
            "Successfully removed {} as child of {}", destination.name, source.name
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )


@relationship_router.put("/collection/{parent_id}/child_of/{child_id}")
@requires(["hippo:admin", "hippo:write"])
async def add_child_collection(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Add a child collection to a parent collection.
    """

    logger.info(
        "Request to add collection {} as child of {} from {}",
        child_id,
        parent_id,
        request.user.display_name,
    )

    try:
        await collection.add_child(
            parent_id=parent_id,
            child_id=child_id,
            groups=request.user.groups,
            scopes=request.auth.scopes,
        )
        logger.info("Successfully added {} as child of {}", child_id, parent_id)
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )


@relationship_router.delete("/collection/{child_id}/child_of/{parent_id}")
@requires(["hippo:admin", "hippo:write"])
async def remove_child_collection(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Remove a parent-child relationship between two collections.
    """

    logger.info(
        "Request to remove collection {} as child of {} from {}",
        child_id,
        parent_id,
        request.user.display_name,
    )

    try:
        await collection.remove_child(
            parent_id=parent_id,
            child_id=child_id,
            groups=request.user.groups,
            scopes=request.auth.scopes,
        )
        logger.info("Successfully removed {} as child of {}", child_id, parent_id)
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )


@relationship_router.post("/collection/{id}/pin")
@requires(["hippo:admin"])
async def pin_collection(
    id: PydanticObjectId,
    request: Request,
) -> PydanticObjectId:
    """
    Pin a collection.
    """

    logger.info("Request to pin collection: {} from {}", id, request.user.display_name)

    coll = await collection.pin(
        id=id, groups=request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Successfully pinned collection {} from {}", id, request.user.display_name
    )
    return coll.id


@relationship_router.post("/collection/{id}/unpin")
@requires(["hippo:admin"])
async def unpin_collection(
    id: PydanticObjectId,
    request: Request,
) -> PydanticObjectId:
    """
    Unpin a collection.
    """

    logger.info(
        "Request to unpin collection: {} from {}", id, request.user.display_name
    )

    coll = await collection.unpin(
        id=id, groups=request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Successfully unpinned collection {} from {}", id, request.user.display_name
    )
    return coll.id



"""
Web frontend for hippo. This consumes the service layer.

NOTE: Code coverage is an explicit NON-goal for the web
      frontend, and as such it is excluded from our
      coverage metrics.
"""

from hashlib import md5
from typing import Literal

from beanie import PydanticObjectId
from fastapi import HTTPException, Request, status
from fastapi.responses import RedirectResponse
from loguru import logger

from hipposerve.database import Collection
from hipposerve.service import collection, product, storage
from hipposerve.settings import SETTINGS

from .auth import router as auth_router
from .router import static_files as static_files
from .router import templates, web_router
from .search import router as search_router

web_router.include_router(search_router)
web_router.include_router(auth_router)


def get_overflow_content(lst: list[Collection], type: Literal["collection", "product"]):
    if len(lst) < 6:
        return None

    overflow_content = ""
    for n in range(6, len(lst) + 1):
        overflow_content += f'<a href="{SETTINGS.web_root}/{type}s/{lst[n - 1].id}">{lst[n - 1].name}</a><br>'

    return overflow_content


def get_color_hash(id: str):
    return f"#{md5(id.encode('utf-8')).hexdigest()[:6]}"


def get_cmap(ids: list[str] = []):
    cmap = {}
    for id in ids:
        cmap[id] = get_color_hash(id)
    return cmap


@web_router.get("/")
async def index(request: Request):
    recent_products = await product.read_most_recent(
        groups=request.user.groups,
        scopes=request.auth.scopes,
        fetch_links=True,
        maximum=16,
    )
    recent_collections = await collection.read_most_recent(
        groups=request.user.groups, scopes=request.auth.scopes, maximum=16
    )
    pinned_products = await product.read_pinned(
        groups=request.user.groups,
        scopes=request.auth.scopes,
        fetch_links=True,
        maximum=16,
    )
    pinned_collections = await collection.read_pinned(
        groups=request.user.groups, scopes=request.auth.scopes, maximum=16
    )
    logger.info(
        "Index page requested by {} with {}/{} recent products/collections and {}/{} "
        "pinned products/collections, user has {} scopes",
        request.user.display_name,
        len(recent_products),
        len(recent_collections),
        len(pinned_products),
        len(pinned_collections),
        request.auth.scopes,
    )
    return templates.TemplateResponse(
        "index.html",
        {
            "request": request,
            "recent_products": recent_products,
            "recent_collections": recent_collections,
            "pinned_products": pinned_products,
            "pinned_collections": pinned_collections,
            "user": request.user.display_name,
            "web_root": SETTINGS.web_root,
        },
    )


@web_router.get("/products/{id}")
async def product_view(request: Request, id: str):
    product_instance = await product.read_by_id(
        id, request.user.groups, scopes=request.auth.scopes
    )
    # Grab the history!
    latest_version = await product.walk_to_current(
        product_instance, request.user.groups, scopes=request.auth.scopes
    )
    version_history = await product.walk_history(latest_version)

    logger.info(
        "Product page for {} requested by {}, user has {} scopes",
        id,
        request.user.display_name,
        request.auth.scopes,
    )

    return templates.TemplateResponse(
        "product.html",
        {
            "request": request,
            "product": product_instance,
            "sources": product_instance.sources,
            "versions": version_history,
            "user": request.user.display_name,
            "web_root": SETTINGS.web_root,
        },
    )


@web_router.get("/products/{id}/edit")
async def product_edit(request: Request, id: str):
    product_instance = await product.read_by_id(
        id, request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Product edit page for {} requested by {}, user has {} scopes",
        id,
        request.user.display_name,
        request.auth.scopes,
    )

    return templates.TemplateResponse(
        "product_edit.html",
        {
            "request": request,
            "product": product_instance,
            "user": request.user.display_name,
            "web_root": SETTINGS.web_root,
        },
    )


@web_router.get("/products/{id}/{slug}")
async def product_read_slug(request: Request, id: str, slug: str) -> RedirectResponse:
    product_instance = await product.read_by_id(
        id, request.user.groups, scopes=request.auth.scopes
    )

    # If they have not 401/403'd from reading the product, they can read sources
    try:
        presigned = await storage.read(
            file=product_instance.sources[slug], storage=request.app.storage
        )
    except (FileNotFoundError, KeyError):
        logger.warning(
            "Slug {} not found for product {}, requested by {}, user has {} scopes",
            slug,
            id,
            request.user.display_name,
            request.auth.scopes,
        )
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Slug {slug} not found for product {id}",
        )

    logger.info(
        "Product source {} for product {} requested by {}, user has {} scopes, redirecting",
        slug,
        id,
        request.user.display_name,
        request.auth.scopes,
    )

    return RedirectResponse(url=presigned, status_code=status.HTTP_302_FOUND)


@web_router.get("/collections/{id}/edit")
async def collection_edit(request: Request, id: PydanticObjectId):
    collection_instance = await collection.read(
        id, request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Collection edit page for {} requested by {}, user has {} scopes",
        id,
        request.user.display_name,
        request.auth.scopes,
    )

    return templates.TemplateResponse(
        "collection_edit.html",
        {
            "request": request,
            "collection": collection_instance,
            "user": request.user.display_name,
            "web_root": SETTINGS.web_root,
        },
    )


@web_router.get("/collections/{id}")
async def collection_view(request: Request, id: PydanticObjectId):
    collection_instance = await collection.read(
        id, request.user.groups, scopes=request.auth.scopes
    )
    parents_overflow_content = get_overflow_content(
        collection_instance.parent_collections, "collection"
    )
    children_overflow_content = get_overflow_content(
        collection_instance.child_collections, "collection"
    )
    cmap = get_cmap(
        [str(id)]
        + [str(x.id) for x in collection_instance.parent_collections]
        + [str(x.id) for x in collection_instance.child_collections]
    )
    logger.info(
        "Collection page for {} requested by {}, user has {} scopes",
        id,
        request.user.display_name,
        request.auth.scopes,
    )
    return templates.TemplateResponse(
        "collection.html",
        {
            "request": request,
            "collection": collection_instance,
            "parents_overflow_content": parents_overflow_content,
            "children_overflow_content": children_overflow_content,
            "cmap": cmap,
            "user": request.user.display_name,
            "web_root": SETTINGS.web_root,
            "api_root": SETTINGS.api_root,
        },
    )



"""
Request and response models for the users endpoints.
"""

from pydantic import BaseModel

from hipposerve.database import ComplianceInformation
from hipposerve.service.users import Privilege


class CreateUserRequest(BaseModel):
    privileges: list[Privilege]
    password: str | None
    compliance: ComplianceInformation | None
    email: str | None
    avatar_url: str | None
    gh_profile_url: str | None


class CreateUserResponse(BaseModel):
    api_key: str | None


class ReadUserResponse(BaseModel):
    name: str
    privileges: list[Privilege]
    compliance: ComplianceInformation | None


class UpdateUserRequest(BaseModel):
    privileges: list[Privilege] | None = None
    refresh_key: bool = False
    password: str | None = None
    compliance: ComplianceInformation | None = None
    add_groups: list[str] | None = None
    remove_groups: list[str] | None = None


class UpdateUserResponse(BaseModel):
    api_key: str | None
    compliance: ComplianceInformation | None = None



"""
The text editor flow for hippo. Uses textual to launch a TUI app and edit
the text associated with a collection or product.
"""

import ast
import json
from pathlib import Path

import xxhash
from beanie import PydanticObjectId
from textual import on
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.reactive import reactive
from textual.widgets import (
    Button,
    Checkbox,
    Footer,
    Input,
    Label,
    RadioButton,
    RadioSet,
    Select,
    TabbedContent,
    TabPane,
    TextArea,
)

from hippometa import ALL_METADATA, ALL_METADATA_TYPE

from . import product
from .core import Client


class EditorApp(App):
    CSS_PATH = "textedit.tcss"
    _product_name = reactive("")
    _description = reactive("")
    _sources = reactive(None)
    _metadata = reactive(None)
    _selected_metadata_class: ALL_METADATA_TYPE | None
    _current_version: str
    _projected_version: str
    _version: int = 2
    _drop_sources = []
    _replace_sources = []
    _new_sources = []

    BINDINGS = [
        ("d", "show_tab('details-tab')", "Details"),
        ("s", "show_tab('sources-tab')", "Sources"),
        ("m", "show_tab('metadata-tab')", "Metadata"),
    ]

    def __init__(self, id: PydanticObjectId, client: Client):
        self._object_id = id
        self._client = client

        _product = product.read(client, id)

        self._current_version = _product.version
        self._projected_version = self.project_version()

        super().__init__()

        self._product_name = _product.name
        self._description = _product.description
        self._sources = _product.sources
        self._metadata = _product.metadata

    def compose(self) -> ComposeResult:
        """
        This function builds the Textual GUI components. The GUI contains a footer
        and a tabbed display. The footer contains keyboard shortcuts to help
        navigate the tabbed display. The available tabs and their descriptions are:
          Details: Enables edits to product title, product description, and version level
          Sources: Enables adding, replacing or deleting sources
          Metadata: Enables changing metadata type and/or editing metadata fields
        """
        yield Footer()

        with TabbedContent():
            # Construct the Details tab
            with TabPane("Details", id="details-tab"):
                with Vertical(classes="title"):
                    yield Label("Product Title:", variant="accent")
                    yield Input(
                        value=self._product_name,
                        placeholder="Product title...",
                        id="product-title",
                    )
                with Vertical(classes="description"):
                    yield Label("Product Description:", variant="accent")
                    yield TextArea(
                        classes="textarea",
                        text=self._description,
                        language="markdown",
                        id="product-description",
                    )
                with Vertical(classes="versioning"):
                    yield Label("Version Bump:", variant="accent")
                    yield RadioSet(
                        RadioButton("Major"),
                        RadioButton("Minor"),
                        RadioButton("Patch", value=True),
                        id="version",
                    )
                    yield Label(
                        f"{self._current_version} → {self._projected_version}",
                        id="versionproject",
                        variant="success",
                    )
                with Horizontal(classes="save-or-cancel"):
                    yield Button("Save", variant="success", id="save")
                    yield Button("Cancel", variant="error", id="cancel")

            # Construct the Sources tab
            with TabPane("Sources", id="sources-tab"):
                for source in self._sources:
                    with Horizontal(classes="sources existing-source"):
                        with Vertical():
                            yield Label("Filename:", variant="accent")
                            yield Input(
                                value=source.name,
                                placeholder="",
                                id=f"edit-source-name-{source.uuid}",
                                tooltip=(
                                    "Enter the absolute filepath to a file of the same name to "
                                    "update the file; leave this field as-is if only editing "
                                    "other fields (i.e., source description)"
                                ),
                            )
                        with Vertical():
                            yield Label("Source Description:", variant="accent")
                            yield Input(
                                value=source.description,
                                placeholder="Description...",
                                id=f"edit-source-desc-{source.uuid}",
                            )
                        with Vertical():
                            yield Label("Is Primary Source?", variant="accent")
                            yield Checkbox(
                                "Primary",
                                classes="primary-source-checkbox",
                                id=f"edit-primary-source-checkbox-{source.uuid}",
                            )
                        with Vertical():
                            yield Label("Delete Source?", variant="accent")
                            yield Checkbox(
                                "Yes",
                                classes="delete-source-checkbox",
                                id=f"delete_source_checkbox_{source.uuid}",
                            )
                with Horizontal(classes="sources new-source"):
                    with Vertical():
                        yield Label("Filepath:", variant="accent")
                        yield Input(
                            value=None,
                            placeholder="Path to new source file...",
                            id="new-source-file-path",
                        )
                    with Vertical():
                        yield Label("Source Description:", variant="accent")
                        yield Input(
                            value=None,
                            placeholder="Source description...",
                            id="new-source-file-desc",
                        )
                    with Vertical():
                        yield Label("Is Primary Source?", variant="accent")
                        yield Checkbox(
                            "Primary",
                            classes="primary-source-checkbox",
                            id="new-primary-source-checkbox",
                        )
                with Horizontal(classes="save-or-cancel"):
                    yield Button("Save", variant="success", id="save")
                    yield Button("Cancel", variant="error", id="cancel")

            # Construct the Metadata tab
            with TabPane("Metadata", id="metadata-tab"):
                class_schema = self._metadata.model_json_schema()
                class_title = class_schema["title"]
                yield Label("Metadata Type:", variant="primary")
                metadata_options = [
                    v.model_json_schema()["title"]
                    for v in ALL_METADATA.values()
                    if v is not None
                ]
                yield Select.from_values(
                    metadata_options, value=class_title, id="metadata-selector"
                )
                with Horizontal(classes="metadata-inputs-container"):
                    None
                with Horizontal(classes="save-or-cancel"):
                    yield Button("Save", variant="success", id="save")
                    yield Button("Cancel", variant="error", id="cancel")

    def action_show_tab(self, tab: str) -> None:
        """Switch to a new tab."""
        self.get_child_by_type(TabbedContent).active = tab

    def project_version(self):
        split = [int(x) for x in self._current_version.split(".")]
        split[self._version] = split[self._version] + 1
        for index in range(self._version + 1, len(split)):
            split[index] = 0
        return ".".join([f"{x}" for x in split])

    def update_projected_version(self):
        """Update the Textual GUI's "Version Bump" string (ie 1.0.0 -> 1.0.1)"""
        self._projected_version = self.project_version()
        version_label = self.query_one("#versionproject")
        version_label.update(f"{self._current_version} → {self._projected_version}")

    @on(RadioSet.Changed)
    def on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        """Handle radio set changes."""
        if event.radio_set.id == "version":
            self._version = {"Major": 0, "Minor": 1, "Patch": 2}[
                event.pressed.label.plain
            ]
            self.update_projected_version()

    @on(Checkbox.Changed)
    def on_checkbox_changed(self, event: Checkbox.Changed) -> None:
        """Handle checkbox changes; currently only 'Delete Source?' field has a checkbox"""
        if event.checkbox.has_class("delete-source-checkbox"):
            split_id = event.checkbox.id.split("_")
            uuid = split_id[len(split_id) - 1]
            toggled_source = [source for source in self._sources if source.uuid == uuid]
            if toggled_source[0].name in self._drop_sources:
                self._drop_sources.remove(toggled_source[0].name)
            else:
                self._drop_sources.append(toggled_source[0].name)
            return

    def generate_metadata_fields(self):
        original_class_schema = self._metadata.model_json_schema()
        original_class_json = self._metadata.model_dump(mode="json")
        original_class_schema = self._metadata.model_json_schema()
        original_class_title = original_class_schema["title"]
        is_original_selected = (
            original_class_title
            == self._selected_metadata_class.model_json_schema()["title"]
        )

        for field_key, field_data in self._selected_metadata_class.model_json_schema()[
            "properties"
        ].items():
            if field_key == "metadata_type":
                continue
            vertical = Vertical()
            self.query_one(".metadata-inputs-container").mount(vertical)
            vertical.mount(Label(field_data["title"], variant="accent"))
            if "enum" in field_data:
                vertical.mount(
                    Select.from_values(
                        field_data["enum"],
                        id=field_key,
                        value=str(original_class_json[field_key])
                        if is_original_selected
                        else Select.BLANK,
                    )
                )
            elif "additionalProperties" in field_data:
                vertical.add_class("metadata-textarea")
                vertical.mount(
                    TextArea.code_editor(
                        classes="textarea",
                        language="json",
                        text=json.dumps(original_class_json[field_key], indent=2)
                        if is_original_selected
                        else "",
                        id=field_key,
                    )
                )
            else:
                vertical.mount(
                    Input(
                        value=str(original_class_json[field_key])
                        if is_original_selected
                        else "",
                        id=field_key,
                        placeholder=f"Enter {field_data['title'].lower()}...",
                    )
                )

    @on(Select.Changed)
    def select_changed(self, event: Select.Changed) -> None:
        if event.select.id == "metadata-selector":
            selected_class_title = event.select.value
            if selected_class_title is Select.BLANK:
                # if _selected_metadata_class is None, revert back to stored _metadata value
                if self._selected_metadata_class is None:
                    class_schema = self._metadata.model_json_schema()
                    class_title = class_schema["title"]
                    event.select.value = class_title
                # if _selected_metadata_class exists, revert back to stored value of _selected_metadata_class
                else:
                    saved_class_schema = (
                        self._selected_metadata_class.model_json_schema()
                    )
                    saved_class_schema_title = saved_class_schema["title"]
                    event.select.value = saved_class_schema_title
            else:
                metadata_class = next(
                    v
                    for v in ALL_METADATA.values()
                    if v.model_json_schema()["title"] == selected_class_title
                )
                self._selected_metadata_class = metadata_class
            self.query_one(".metadata-inputs-container").remove_children(Vertical)
            self.query_one(".metadata-inputs-container").remove_children(Label)
            self.query_one(".metadata-inputs-container").remove_children(Input)
            self.query_one(".metadata-inputs-container").remove_children(Select)
            self.generate_metadata_fields()

    def get_file_metadata(self, path, source_description):
        source = Path(path)
        with source.open("rb") as file:
            file_info = {
                "name": source.name,
                "size": source.stat().st_size,
                "checksum": f"xxh64:{xxhash.xxh64(file.read()).hexdigest()}",
                "description": source_description,
            }
            return file_info

    def get_new_source(self):
        file_path = self.query_one("#new-source-file-path").value
        source_desc = self.query_one("#new-source-file-desc").value
        if file_path:
            new_source_metadata = self.get_file_metadata(file_path, source_desc)
            return [new_source_metadata]
        else:
            return []

    def get_replace_sources(self):
        source_data = []
        for source in self._sources:
            source_name = self.query_one(f"#edit-source-name-{source.uuid}").value
            source_desc = self.query_one(f"#edit-source-desc-{source.uuid}").value
            is_set_to_delete = self.query_one(
                f"#delete_source_checkbox_{source.uuid}"
            ).value
            if source_name == source.name and is_set_to_delete is False:
                source_metadata = {
                    "name": source.name,
                    "size": source.size,
                    "checksum": source.checksum,
                    "description": source_desc,
                }
                source_data.append(source_metadata)
                continue
            if source_name != source.name and is_set_to_delete is False:
                source_metadata = self.get_file_metadata(source_name, source_desc)
                source_data.append(source_metadata)
        return source_data

    def get_metadata_type(self, field_data):
        if "anyOf" in field_data:
            options = [x for x in field_data["anyOf"]]
            true_type = [x for x in options if x["type"] != "null"][0]
            return {
                "true_type": true_type["type"],
                "is_optional": ({"type": "null"} in options),
            }
        elif "enum" in field_data:
            return {"true_type": "enum", "is_optional": False}
        elif "additionalProperties" in field_data:
            return {
                "true_type": field_data["type"],
                "is_optional": False,
                "empty_value": {},
            }
        return {"true_type": "string", "is_optional": False}

    def get_metadata_changes(self):
        metadata_schema = (
            self._metadata.model_json_schema()
            if self._selected_metadata_class is None
            else self._selected_metadata_class.model_json_schema()
        )
        metadata_properties = metadata_schema["properties"]
        metadata_key_name = metadata_properties["metadata_type"]["default"]
        new_metadata_values = (
            self._metadata.model_dump(mode="json")
            if type(self._metadata) is self._selected_metadata_class
            else {}
        )
        for field_key, field_data in metadata_properties.items():
            if field_key == "metadata_type":
                new_metadata_values[field_key] = metadata_key_name
                continue
            metadata_type = self.get_metadata_type(field_data)
            input_field = self.query_one(f"#{field_key}")
            input_field_value = (
                input_field.text
                if input_field.has_class("textarea")
                else input_field.value
            )
            if "additionalProperties" in field_data:
                new_metadata_values[field_key] = (
                    metadata_type["empty_value"]
                    if bool(input_field_value) is False
                    else ast.literal_eval(input_field_value)
                )
                continue
            if input_field_value == "None" and metadata_type["is_optional"] is True:
                new_metadata_values[field_key] = None
                continue
            if metadata_type["true_type"] == "array":
                new_metadata_values[field_key] = (
                    None
                    if bool(input_field_value) is False
                    and metadata_type["is_optional"] is True
                    else input_field_value.split(",")
                )
                continue
            if metadata_type["true_type"] == "number":
                new_metadata_values[field_key] = (
                    None
                    if bool(input_field_value) is False
                    and metadata_type["is_optional"] is True
                    else int(input_field_value)
                )
                continue
            new_metadata_values[field_key] = (
                None
                if bool(input_field_value) is False
                and metadata_type["is_optional"] is True
                else input_field_value
            )
        return ALL_METADATA[metadata_key_name](**new_metadata_values)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        if event.button.id == "save":
            product.update(
                self._client,
                self._object_id,
                self.query_one("#product-title").value,
                self.query_one("#product-description").text,
                self._version,
                self.get_metadata_changes().model_dump(),
                self.get_new_source(),
                self.get_replace_sources(),
                self._drop_sources,
            )
            exit(0)
        elif event.button.id == "cancel":
            exit(0)
        else:
            raise AttributeError(f"Unknown button pressed: {event.button.id}")


def edit_product(client: Client, id: PydanticObjectId):
    """
    Edit the text of a product. This downloads the metadata for the product,
    and then launches a text editor to edit the text. Once the text is saved,
    the updated metadata is uploaded (if changed).
    """

    app = EditorApp(id, client)
    app.run()



"""
Builds the router and templates.
"""

from pathlib import Path

from fastapi import APIRouter
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from jinja2 import Environment, PackageLoader

from hipposerve.settings import SETTINGS

web_router = APIRouter(prefix="/web")

potential_routes = ["hipposerve/web", "hippo/hipposerve/web", Path(__file__).parent]


def useful_variables(_):
    return {
        "soauth_enabled": SETTINGS.auth_system == "soauth",
        "soauth_login_url": f"{SETTINGS.soauth_service_url}/login/{SETTINGS.soauth_app_id}",
    }


for route in potential_routes:
    try:
        templates = Jinja2Templates(
            env=Environment(
                loader=PackageLoader("hipposerve", package_path="web/templates"),
                extensions=["jinja_markdown.MarkdownExtension"],
            ),
            context_processors=[useful_variables],
        )

        static_files = {
            "path": "/web/static",
            "name": "static",
            "app": StaticFiles(directory=f"{route}/static", follow_symlink=True),
        }

        break
    except (FileNotFoundError, RuntimeError):
        continue


if "templates" not in locals():
    raise FileNotFoundError("Could not find the templates directory.")

if "static_files" not in locals():
    raise FileNotFoundError("Could not find the static directory.")



"""
Server settings, uses pydantic settings models.
"""

from pwdlib import PasswordHash
from pwdlib.hashers.argon2 import Argon2Hasher
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    minio_url: str
    minio_presign_url: str | None = None
    minio_upgrade_presign_url_to_https: bool = False
    minio_secure: bool = False
    minio_cert_check: bool = False

    minio_access: str
    minio_secret: str

    mongo_uri: str

    title: str
    description: str
    api_root: str = ""

    hasher: PasswordHash = PasswordHash([Argon2Hasher()])

    add_cors: bool = True
    debug: bool = True

    web: bool = False
    "Serve the web frontend."
    web_root: str = "/web"

    auth_system: str | None = None
    soauth_service_url: str | None = None
    soauth_app_id: str | None = None
    soauth_public_key: str | None = None
    soauth_base_url: str | None = None
    soauth_client_secret: str | None = None
    soauth_key_pair_type: str = "Ed25519"


SETTINGS = Settings()



"""
Access control setup (changing access control levels, checking
for access to documents). Works with `ProtectedDocument`s.
"""

from typing import Iterable

from hipposerve.database import ProtectedDocument
from hipposerve.service import versioning
from hipposerve.service.auth import AuthenticationError


async def update_access_control(
    doc: ProtectedDocument,
    owner: str | None = None,
    add_readers: list[str] | None = None,
    remove_readers: list[str] | None = None,
    add_writers: list[str] | None = None,
    remove_writers: list[str] | None = None,
) -> ProtectedDocument:
    """
    Update the access control on this doc, either adding or removing
    readers or writers. Note that access control changes walk the entire
    tree and update all docs, and do not create a new version.
    """

    initial_doc_id = doc.id

    if not doc.current:
        raise versioning.VersioningError(
            "Attempting to update a non-current product. You must always "
            "make changes to the head of the list"
        )

    readers = (set(doc.readers) | set(add_readers or [])) - set(remove_readers or [])
    writers = (set(doc.writers) | set(add_writers or [])) - set(remove_writers or [])

    if owner:
        readers.add(owner)
        writers.add(owner)
    else:
        owner = doc.owner

    # Need to walk the tree; some protected resources have versions.

    while doc.replaces is not None:
        doc.readers = readers
        doc.writers = writers
        doc.owner = owner

        await doc.save()

        doc = doc.replaces

        if not hasattr(doc, "replaces"):
            link = doc
            doc = await link.fetch()

    # Base of tree that has no replacement
    doc.readers = readers
    doc.writers = writers
    doc.owner = owner

    await doc.save()

    # We use the classmethod on the _instance_ to get back the same
    # type that the doc is! Using ProtectedDocument.get() doesn't work.
    return await doc.get(initial_doc_id)


def check_user_access(
    user_groups: Iterable[str],
    document_groups: Iterable[str],
    scopes: Iterable[str] = set(),
) -> bool:
    """
    Check whether a user (based on their groups) has access to a specific
    document based on its 'reader' or 'writer' group list (`document_groups`).

    Note that the group `admin` is always added to `allowed`.

    Raises
    ------
    AuthenticationError
        When there is no overlap between `user_groups` and `document_groups`
    """

    if "hippo:admin" in scopes:
        return True

    allowed = set(document_groups)
    user_groups = set(user_groups)

    overlap = user_groups & allowed

    if overlap:
        return True

    raise AuthenticationError(
        "User does not have the required group access for this operation"
    )



# from abc import ABC, abstractmethod
from pathlib import Path

import httpx
from pydantic import BaseModel, Field
from rich.console import Console

from henry.source import LocalSource, RemoteSource
from hippoclient import product as product_client
from hippoclient.caching import MultiCache
from hippoclient.product import ProductMetadata
from hippometa import ALL_METADATA_TYPE

from .exceptions import InvalidSlugError, PreflightFailedError, ProductIncompleteError


class ProductInstance(BaseModel):
    pass

    def _upload(
        self,
        client: httpx.Client,
        console: Console,
        skip_preflight: bool = False,
        readers: list[str] | None = None,
        writers: list[str] | None = None,
    ) -> str:
        return


class LocalProduct(ProductInstance):
    """
    A local product, created before pushing up to HIPPO. This includes references
    to local files that are not yet ingested into the HIPPO system.
    """

    product_id: str | None = None
    name: str
    description: str
    metadata: ALL_METADATA_TYPE
    sources: dict[str, LocalSource] = {}
    readers: list[str] = []
    writers: list[str] = []

    def model_post_init(self, __context):
        for k, v in self.sources.items():
            v.slug = k
            self[k] = v

        return super().model_post_init(__context)

    def __handle_key_error(self, key: str):
        if key in self.metadata.valid_slugs:
            raise InvalidSlugError(
                f"Slug {key} is valid for this metadata type, but not present in this object"
            )
        else:
            raise InvalidSlugError(
                f"Slug {key} not in valid list: {self.metadata.valid_slugs}"
            )

    def keys(self):
        return self.sources.keys()

    def values(self):
        return self.sources.values()

    def items(self):
        return self.sources.items()

    def get(self, key, default=None, /):
        return self.sources.get(key, default)

    def clear(self):
        return self.sources.clear()

    def setdefault(self, key, default=None, /):
        if key not in self.metadata.valid_slugs:
            self.__handle_key_error(key=key)

        return self.sources.setdefault(key, default)

    def pop(self, key):
        try:
            return self.sources.pop(key)
        except KeyError:
            self.__handle_key_error(key=key)

    def popitem(self):
        try:
            return self.sources.popitem()
        except KeyError:
            raise InvalidSlugError("No slugs left to pop (gross!)")

    def copy(self):
        new = LocalProduct(
            name=self.name, description=self.description, metadata=self.metadata
        )

        new.sources = self.sources.copy()

        return new

    def __len__(self) -> int:
        return self.sources.__len__()

    def __getitem__(self, key):
        try:
            return self.sources.__getitem__(key)
        except KeyError:
            self.__handle_key_error(key=key)

    def __setitem__(self, key, value):
        if key not in self.metadata.valid_slugs:
            self.__handle_key_error(key=key)

        if isinstance(value, LocalSource):
            self.sources.__setitem__(key, value)
        elif isinstance(value, (str, Path)):
            self.sources.__setitem__(
                key, LocalSource(path=Path(value), slug=key, description=None)
            )
        else:
            raise TypeError(
                f"Value for key '{key}' must be a LocalSource, str, or Path, not {type(value).__name__}"
            )

    def __delitem__(self, key):
        try:
            self.sources.__delitem__(key)
        except KeyError:
            self.__handle_key_error(key=key)

    def __missing__(self, key):
        self.__handle_key_error(key=key)

    def __iter__(self):
        return self.sources.__iter__()

    def __reversed__(self):
        return self.sources.__reversed__()

    def __contains__(self, key):
        return self.sources.__contains__(key)

    def __repr__(self):
        return (
            f"LocalProduct(name='{self.name}', description='{self.description}', "
            f"metadata={self.metadata.__repr__()}, "
            f"{', '.join([f'{x}={y.__repr__()}' for x, y in self.sources.items()])})"
        )

    def __str__(self):
        return (
            f"{self.name} ({self.description}) -- {self.metadata.__str__()} "
            f"-- Sources: {', '.join([x.__str__() for x in self.sources.values()])}"
        )

    def preflight(self):
        """
        Run a pre-flight check on this product. This checks that:

        a) The product has a valid name and description.
        b) That each source has a valid slug, name, and description.
        c) That each source's path points to an existing valid file.

        If any check fails, we raise a PreflightFailedError.
        """

        if self.product_id:
            # We've already flown!
            return

        if len(self.name) < 2 or not isinstance(self.name, str):
            raise PreflightFailedError(
                f"Name: {self.name} is not a valid name; ensure it is at least 2 characters and a valid string"
            )

        if len(self.description) < 2 or not isinstance(self.description, str):
            raise PreflightFailedError(
                f"Description: {self.description} is not a valid description; ensure it is at least 2 characters and a valid string"
            )

        for slug, source in self.sources.items():
            if slug not in self.metadata.valid_slugs:
                raise PreflightFailedError(
                    f"Slug {slug} not valid for upload for metadata type {type(self.metadata).__name__}"
                )
            if source.slug not in self.metadata.valid_slugs:
                raise PreflightFailedError(
                    f"Slug {source.slug} not valid for upload for metadata type {type(self.metadata).__name__}"
                )
            if (
                not source.description
                or len(source.description) < 2
                or not isinstance(source.description, str)
            ):
                raise PreflightFailedError(
                    f"Description for source {source.slug} is not valid; ensure it is at least 2 characters and a valid string"
                )
            if len(source.name) < 2 or not isinstance(source.name, str):
                raise PreflightFailedError(
                    f"Name for source {source.slug} is not valid; ensure it is at least 2 characters and a valid string"
                )
            if not source.path.exists():
                raise PreflightFailedError(
                    f"Source {source.slug} has non-existent source file at {source.path}"
                )

        return

    def _upload(
        self,
        client: httpx.Client,
        console: Console,
        skip_preflight: bool = False,
        readers: list[str] | None = None,
        writers: list[str] | None = None,
    ) -> str:
        if self.product_id:
            # We've already been uploaded
            return self.product_id

        if not skip_preflight:
            self.preflight()

        remote_id = product_client.create(
            client=client,
            name=self.name,
            description=self.description,
            metadata=self.metadata,
            sources={x: y.path for x, y in self.sources.items()},
            source_descriptions={x: y.description for x, y in self.sources.items()},
            readers=readers,
            writers=writers,
            console=console,
        )
        self.product_id = remote_id
        return remote_id


class RemoteProduct(ProductInstance):
    """
    A remote product, created by pulling down from HIPPO. This includes references
    to 'remote' files - either those in a cache or in-memory after download.
    """

    product_id: str
    name: str
    description: str
    metadata: ALL_METADATA_TYPE
    version: str | None = None
    sources: dict[str, RemoteSource] = {}
    readers: list[str] = []
    writers: list[str] = []

    def model_post_init(self, __context):
        for k, v in self.sources.items():
            v.slug = k

        return super().model_post_init(__context)

    def __handle_key_error(self, key: str):
        if key in self.sources.valid_slugs:
            raise InvalidSlugError(
                f"Slug {key} is valid for this metadata type, but not present in this object"
            )
        else:
            raise InvalidSlugError(
                f"Slug {key} not in valid list: {self.metadata.valid_slugs}"
            )

    @classmethod
    def pull(
        cls,
        product_id: str,
        client: httpx.Client,
        cache: MultiCache,
        console: Console,
        realize_sources: bool = True,
    ) -> "RemoteProduct":
        product_metadata = product_client.read(
            client=client, id=product_id, console=console
        )

        if realize_sources:
            # Ensure it's cached
            product_client.cache(
                client=client,
                cache=cache,
                id=product_id,
                console=console,
            )

        # Convert the sources
        sources = {
            x: RemoteSource(
                slug=x,
                name=y.name,
                description=y.description,
                source_id=y.uuid,
                cache=cache,
                realize=realize_sources,
            )
            for x, y in product_metadata.sources.items()
        }

        return cls(
            product_id=product_id,
            name=product_metadata.name,
            description=product_metadata.description,
            metadata=product_metadata.metadata,
            sources=sources,
            version=product_metadata.version,
            readers=product_metadata.readers,
            writers=product_metadata.writers,
        )

    @classmethod
    def read(cls, directory: Path | str, allow_incomplete: bool) -> "RemoteProduct":
        """
        Read a product that was serialized to disk, usually with the
        `henry prodcut download $ID` command.
        """
        directory = Path(directory)

        with open(directory / "product.json", "r") as handle:
            core_metadata = ProductMetadata.model_validate_json(handle.read())

        sources = {}

        for slug, source_metadata in core_metadata.sources.items():
            path = directory / slug / source_metadata.name
            exists = path.exists()

            if (not allow_incomplete) and (not exists):
                raise ProductIncompleteError(
                    "This product is incomplete, try redownloading or set allow_incomplete=True"
                )
            else:
                sources[slug] = RemoteSource(
                    path=path,
                    slug=slug,
                    name=source_metadata.name,
                    description=source_metadata.description,
                    source_id=source_metadata.uuid,
                    cached=exists,
                    cache=None,
                    realize=False,
                )

        return cls(
            product_id=str(core_metadata.id),
            name=core_metadata.name,
            description=core_metadata.description,
            metadata=core_metadata.metadata,
            sources=sources,
            readers=core_metadata.readers,
            writers=core_metadata.writers,
        )

    def keys(self):
        return self.sources.keys()

    def values(self):
        return self.sources.values()

    def items(self):
        return self.sources.items()

    def get(self, key, default=None, /):
        return self.sources.get(key, default)

    def copy(self):
        new = RemoteProduct(
            name=self.name, description=self.description, metadata=self.metadata
        )

        new.sources = self.sources.copy()

        return new

    def __len__(self) -> int:
        return self.sources.__len__()

    def __getitem__(self, key):
        try:
            return self.sources.__getitem__(key)
        except KeyError:
            self.__handle_key_error(key=key)

    def __missing__(self, key):
        self.__handle_key_error(key=key)

    def __iter__(self):
        return self.sources.__iter__()

    def __reversed__(self):
        return self.sources.__reversed__()

    def __contains__(self, key):
        return self.sources.__contains__(key)

    def __repr__(self):
        return (
            f"RemoteProduct(name='{self.name}', description='{self.description}', "
            f"metadata={self.metadata.__repr__()}, "
            f"{', '.join([f'{x}={y.__repr__()}' for x, y in self.sources.items()])})"
        )

    def __str__(self):
        return (
            f"{self.name} ({self.description}) -- {self.metadata.__str__()} "
            f"-- Sources: {', '.join([x.__str__() for x in self.sources.values()])}"
        )

    def create_revision(
        self, *, major: bool = False, minor: bool = False, patch: bool = False
    ) -> "RevisionProduct":
        """
        Create a revision of this product. Major, minor, and patch are mutually exclusive
        """

        if sum([major, minor, patch]) != 1:
            raise ValueError("You must choose exactly _one_ of major, minor, or patch")

        if major:
            revision_level = 0
        elif minor:
            revision_level = 1
        else:
            revision_level = 2

        return RevisionProduct(
            revision_of=self,
            revision_level=revision_level,
            readers=self.readers,
            writers=self.writers,
        )

    def preflight(self):
        """
        Preflight pass-through; used when we are adding a collection.
        """
        assert self.product_id, "Cannot preflight a RemoteProduct without a product_id"

    def _upload(self, *args, **kwargs) -> str:
        """
        Upload is not supported for RemoteProduct, as it is a read-only object.
        """
        return self.product_id


class RevisionProduct(ProductInstance):
    """
    A revision of a product, created from a RemoteProduct. Everything starts
    out empty, then if you make changes to this we push those changes to
    the upstream.
    """

    product_id: str | None = None
    revision_of: RemoteProduct

    revision_level: int = Field(ge=0, le=2)
    name: str | None = None
    description: str | None = None
    metadata: ALL_METADATA_TYPE | None = None
    sources: dict[str, LocalSource] = {}

    readers: list[str]
    writers: list[str]

    def model_post_init(self, __context):
        for k, v in self.sources.items():
            v.slug = k
            self[k] = v

        return super().model_post_init(__context)

    def __handle_key_error(self, key: str):
        if key in self.revision_of.metadata.valid_slugs:
            raise InvalidSlugError(
                f"Slug {key} is valid for this metadata type, but not present in this object"
            )
        else:
            raise InvalidSlugError(
                f"Slug {key} not in valid list: {self.revision_of.metadata.valid_slugs}"
            )

    def __setattr__(self, name, value):
        # Cannot change metadata type.
        if name == "metadata" and hasattr(self, "metadata"):
            current = self.revision_of.metadata
            if not isinstance(value, type(current)):
                raise TypeError(
                    f"Cannot change metadata {type(value)} to metadata of type {type(current)}"
                )
        super().__setattr__(name, value)

    def keys(self):
        return self.sources.keys()

    def values(self):
        return self.sources.values()

    def items(self):
        return self.sources.items()

    def get(self, key, default=None, /):
        return self.sources.get(key, default)

    def clear(self):
        return self.sources.clear()

    def setdefault(self, key, default=None, /):
        if key not in self.revision_of.metadata.valid_slugs:
            self.__handle_key_error(key=key)

        return self.sources.setdefault(key, default)

    def pop(self, key):
        try:
            return self.sources.pop(key)
        except KeyError:
            self.__handle_key_error(key=key)

    def popitem(self):
        try:
            return self.sources.popitem()
        except KeyError:
            raise InvalidSlugError("No slugs left to pop (gross!)")

    def copy(self):
        new = RevisionProduct(
            revision_of=self.revision_of,
            name=self.name,
            description=self.description,
            metadata=self.metadata,
        )

        new.sources = self.sources.copy()

        return new

    def __len__(self) -> int:
        return self.sources.__len__()

    def __getitem__(self, key):
        try:
            return self.sources.__getitem__(key)
        except KeyError:
            self.__handle_key_error(key=key)

    def __setitem__(self, key, value):
        if key not in self.revision_of.metadata.valid_slugs:
            self.__handle_key_error(key=key)

        if isinstance(value, LocalSource):
            self.sources.__setitem__(key, value)
        elif isinstance(value, (str, Path)):
            self.sources.__setitem__(
                key, LocalSource(path=Path(value), slug=key, description=None)
            )
        else:
            raise TypeError(
                f"Value for key '{key}' must be a LocalSource, str, or Path, not {type(value).__name__}"
            )

    def __delitem__(self, key):
        try:
            self.sources.__delitem__(key)
        except KeyError:
            self.__handle_key_error(key=key)

    def __missing__(self, key):
        self.__handle_key_error(key=key)

    def __iter__(self):
        return self.sources.__iter__()

    def __reversed__(self):
        return self.sources.__reversed__()

    def __contains__(self, key):
        return self.sources.__contains__(key)

    def __str__(self):
        diff = self._calculate_diff()

        result = ""

        for x, y in diff.items():
            if not y:
                continue

            match y:
                case str():
                    result += (
                        f"{x.capitalize()}: '{getattr(self.revision_of, x)}' -> '{y}'\n"
                    )
                case dict():
                    result += (
                        f"{x.replace('_', ' ').capitalize()}: {', '.join(y.keys())}\n"
                    )
                case BaseModel():
                    left = getattr(self.revision_of, x).model_dump()
                    right = getattr(self, x).model_dump()
                    result += f"{x.capitalize()} diff: "
                    for key in left:
                        if left[key] != right[key]:
                            result += f"[{key}] {left[key]} -> {right[key]}  "
                    result += "\n"

        if not result:
            result = "A revision object with no changes relative to the parent product"
        else:
            result = result.strip()

        return result

    def _calculate_diff(self) -> dict:
        """
        Calculates the diff between this and the RemoteProduct for uploading.
        """

        diff = {}

        for x in ["name", "description", "metadata"]:
            if getattr(self, x) != getattr(self.revision_of, x):
                diff[x] = getattr(self, x)
            else:
                diff[x] = None

        new_sources = {}
        replace_sources = {}

        for source in self:
            if source in self.revision_of:
                replace_sources[source] = self[source]
            else:
                new_sources[source] = self[source]

        diff["new_sources"] = new_sources
        diff["replace_sources"] = replace_sources

        diff.update(
            self._claculate_reader_writer_diff(
                readers=self.readers, writers=self.writers
            )
        )

        return diff

    def _claculate_reader_writer_diff(
        self, readers: list[str], writers: list[str]
    ) -> dict:
        """
        Calculates a difference in readers and writers, if they are provided
        """

        add_readers = set(self.revision_of.readers) - set(readers or [])
        add_writers = set(self.revision_of.writers) - set(writers or [])

        remove_readers = set(readers or []) - set(self.revision_of.readers)
        remove_writers = set(writers or []) - set(self.revision_of.writers)

        diff = {
            "add_readers": add_readers or None,
            "add_writers": add_writers or None,
            "remove_readers": remove_readers or None,
            "remove_writers": remove_writers or None,
        }

        return diff

    def preflight(self):
        diff = self._calculate_diff()

        if self.product_id:
            # We've already flown!
            return

        if not any(diff.values()):
            raise PreflightFailedError(
                "No changes detected in this revision; nothing to upload"
            )

        if name := diff.get("name", None):
            if len(name) < 2 or not isinstance(name, str):
                raise PreflightFailedError(
                    f"Name: {name} is not a valid name; ensure it is at least 2 characters and a valid string"
                )

        if description := diff.get("description", None):
            if len(description) < 2 or not isinstance(description, str):
                raise PreflightFailedError(
                    f"Description: {description} is not a valid description; ensure it is at least 2 characters and a valid string"
                )

        for slug, source in self.sources.items():
            if slug not in self.revision_of.metadata.valid_slugs:
                raise PreflightFailedError(
                    f"Slug {slug} not valid for upload for metadata type {type(self.revision_of.metadata).__name__}"
                )
            if source.slug not in self.revision_of.metadata.valid_slugs:
                raise PreflightFailedError(
                    f"Slug {source.slug} not valid for upload for metadata type {type(self.revision_of.metadata).__name__}"
                )
            if (
                not source.description
                or len(source.description) < 2
                or not isinstance(source.description, str)
            ):
                raise PreflightFailedError(
                    f"Description for source {source.slug} is not valid; ensure it is at least 2 characters and a valid string"
                )
            if len(source.name) < 2 or not isinstance(source.name, str):
                raise PreflightFailedError(
                    f"Name for source {source.slug} is not valid; ensure it is at least 2 characters and a valid string"
                )
            if not source.path.exists():
                raise PreflightFailedError(
                    f"Source {source.slug} has non-existent source file at {source.path}"
                )
        return

    def _upload(
        self,
        client: httpx.Client,
        console: Console,
        skip_preflight: bool = False,
        readers: list[str] | None = None,
        writers: list[str] | None = None,
    ) -> str:
        # Readers/writers are not used!

        if self.product_id:
            # We've already been uploaded
            return self.product_id

        if not skip_preflight:
            self.preflight()

        diff = self._calculate_diff()

        # Break out new and replace sources into paths and descriptions
        new_sources = diff.pop("new_sources", {})
        replace_sources = diff.pop("replace_sources", {})

        diff["new_sources"] = {x: y.path for x, y in new_sources.items()}
        diff["replace_sources"] = {x: y.path for x, y in replace_sources.items()}
        diff["new_source_descriptions"] = {
            x: y.description for x, y in new_sources.items()
        }
        diff["replace_source_descriptions"] = {
            x: y.description for x, y in replace_sources.items()
        }

        remote_id = product_client.update(
            client=client,
            id=self.revision_of.product_id,
            level=self.revision_level,
            **diff,
            console=console,
        )

        self.product_id = remote_id

        return remote_id



"""
Utilities for product search and rendering search results.
"""

from fastapi import APIRouter, Request
from fastapi.responses import HTMLResponse
from loguru import logger

from hippometa import ALL_METADATA
from hipposerve.service import collection, product
from hipposerve.settings import SETTINGS

from .router import templates

router = APIRouter()


# Query and render search results from the navigation bar's "search by name" option
@router.get("/search/results", response_class=HTMLResponse)
async def search_results_view(
    request: Request,
    q: str = None,
    filter: str = "products",
):
    if filter == "products":
        results = await product.search_by_name(
            q, request.user.groups, scopes=request.auth.scopes
        )
    elif filter == "collections":
        results = await collection.search_by_name(
            q, request.user.groups, scopes=request.auth.scopes
        )
    else:
        collection_results = await collection.search_by_owner(
            q, request.user.groups, scopes=request.auth.scopes
        )
        product_results = await product.search_by_owner(
            q, request.user.groups, scopes=request.auth.scopes
        )
        results = collection_results + product_results

    logger.info(
        "Search results for query '{}' with filter '{}' requested by {}, found {} results, user has {} scopes",
        q,
        filter,
        request.user.display_name,
        len(results),
        request.auth.scopes,
    )

    return templates.TemplateResponse(
        "search_results.html",
        {
            "request": request,
            "query": q,
            "filter": filter,
            "results": results,
            "user": request.user.display_name,
            "web_root": SETTINGS.web_root,
        },
    )


# Query and render search results for the more complex metadata request
@router.get("/searchmetadata/results", response_class=HTMLResponse)
async def searchmetadata_results_view(
    request: Request,
    q: str = None,
    filter_on: str = "products",
):
    query_params = dict(request.query_params)

    # Determine which metadata_class we're searching on so we can get the fields;
    # we will use the fields with the query_params to craft type-specific queries
    metadata_class = next(
        v
        for v in ALL_METADATA.values()
        if v.schema()["title"] == query_params["metadata_type"]
    )
    metadata_fields = metadata_class.model_json_schema()["properties"]

    # Create a dict for the following pre-processing of metadata_fields
    simplified_metadata_fields = {}

    # Do some pre-processing of the metadata_fields to simplify query logic
    for field_key, field_data in metadata_fields.items():
        if field_key == "metadata_type" or "additionalProperties" in field_data:
            continue

        if "anyOf" in field_data:
            true_type = [x for x in field_data["anyOf"] if x["type"] != "null"][0]
            field_data = {**field_data, **true_type}
            field_data.pop("anyOf")

        simplified_metadata_fields[field_key] = field_data

    # Create a dict to hold the queries as we craft them below
    metadata_filters = {}

    for key, value in query_params.items():
        if key == "metadata_type":
            continue

        field = simplified_metadata_fields[key]
        field_type = field.get("type", None)

        # Query for enums; no query logic needed
        if "enum" in field:
            metadata_filters[key] = value
        # Query for comma-separated lists
        elif field_type == "array":
            metadata_filters[key] = {"$in": [v.strip() for v in value.split(",")]}
        # Query for number ranges
        elif field_type == "number" and "," in value:
            min, max = value.split(",")
            if min != "undefined":
                metadata_filters[key] = {"$gte": min}
            if max != "undefined":
                metadata_filters[key] = metadata_filters.get(key, {})
                metadata_filters[key]["$lte"] = max
        # Default query applies regex and case insensitivity
        else:
            metadata_filters[key] = {"$regex": value, "$options": "i"}

    results = await product.search_by_metadata(
        metadata_filters, request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Metadata search results for query '{}' with filter '{}' on metadata type '{}' requested by {}, found "
        "{} results, user has {} scopes",
        q,
        filter_on,
        query_params["metadata_type"],
        request.user.display_name,
        len(results),
        request.auth.scopes,
    )

    return templates.TemplateResponse(
        "search_results.html",
        {
            "request": request,
            "query": q,
            "filter": filter_on,
            "results": results,
            "metadata_filters": metadata_filters,
            "metadata_type": query_params["metadata_type"],
            "user": request.user.display_name,
            "web_root": SETTINGS.web_root,
        },
    )


@router.get("/searchmetadata", response_class=HTMLResponse)
async def search_metadata_view(
    request: Request,
):
    metadata_info = {}

    for metadata_class in ALL_METADATA.values():
        if metadata_class is None:
            continue

        class_schema = metadata_class.model_json_schema()
        class_name = class_schema["title"]
        class_fields = class_schema["properties"]

        # Separate number fields from other types so we can render the number
        # fields together in the metadata search page
        number_fields = {}
        other_fields = {}

        for field_key, field_data in class_fields.items():
            # additionalProperties means the field is a dictionary with arbitrary keys
            # we can't search on that.
            if field_key == "metadata_type" or "additionalProperties" in field_data:
                continue

            if "anyOf" in field_data:
                # This occurs when we have x | None, extract x and stick it in the field_data
                true_type = [x for x in field_data["anyOf"] if x["type"] != "null"][0]
                field_data = {**field_data, **true_type}
                field_data.pop("anyOf")

            field_type = field_data.get("type", None)

            if "enum" in field_data:
                other_fields[field_key] = {
                    "type": "enum",
                    "title": field_data["title"],
                    "options": field_data["enum"],
                }
            elif field_type == "array":
                other_fields[field_key] = {
                    "type": "array",
                    "list_arg": field_data["items"]["type"],
                    "title": field_data["title"],
                }
            elif field_type == "number":
                number_fields[field_key] = {
                    "type": "number",
                    "title": field_data["title"],
                    "min": field_data.get("min", None),
                    "max": field_data.get("max", None),
                }
            else:
                other_fields[field_key] = {
                    "type": field_type,
                    "title": field_data["title"],
                }

        metadata_info[class_name] = {**number_fields, **other_fields}

    logger.info(
        "Search metadata page requested by {}, user has {} scopes",
        request.user.display_name,
        request.auth.scopes,
    )

    return templates.TemplateResponse(
        "search_metadata.html",
        {
            "request": request,
            "metadata": metadata_info,
            "user": request.user.display_name,
            "web_root": SETTINGS.web_root,
        },
    )



"""
A CLI interface to the hippo client.
"""

from pathlib import Path
from typing import Annotated

import rich
import typer

import hippoclient as sc

from . import helper
from .core import ClientSettings, MultiCache
from .textedit import edit_product

CLIENT: sc.Client
CACHE: MultiCache
CONSOLE: rich.console.Console

# Meta-setup
APP = typer.Typer()
product_app = typer.Typer(help="Commands for dealing directly with products")
APP.add_typer(product_app, name="product")
collection_app = typer.Typer(help="Commands for dealing with collections")
APP.add_typer(collection_app, name="collection")
cache_app = typer.Typer(
    help="Maintenance commands for the cache. There are also tools for cache management in the product and collection commands."
)
APP.add_typer(cache_app, name="cache")
dev_app = typer.Typer(
    help="Developer commands, mainly used for running and testing servers during hippo development. For regular use, you can ignore these."
)
APP.add_typer(dev_app, name="dev")


@product_app.command("read")
def product_read(id: str):
    """
    Read the information of a product by its ID. You can find the relationship
    between product names and IDs through the product search command.
    """
    global CLIENT, CONSOLE

    product = sc.product.read_with_versions(client=CLIENT, id=id, console=CONSOLE)

    product_extracted_version = product.versions[product.requested]

    CONSOLE.print(product_extracted_version.name, style="bold underline color(3)")
    CONSOLE.print(
        "\nVersions: "
        + helper.render_version_list(
            product.versions, product.current, product.requested
        )
    )
    CONSOLE.print(
        rich.markdown.Markdown(product_extracted_version.description.strip("\n"))
    )
    CONSOLE.print(product_extracted_version.metadata)
    CONSOLE.print(helper.render_source_list(product_extracted_version.sources, CACHE))
    CONSOLE.print("\n" + "Relationships" + "\n", style="bold color(2)")
    CONSOLE.print(
        "Collections: "
        + ", ".join(str(c) for c in product_extracted_version.collections)
    )
    if len(product_extracted_version.parent_of) > 0:
        CONSOLE.print("Children: " + ", ".join(product_extracted_version.parent_of))
    if len(product_extracted_version.child_of) > 0:
        CONSOLE.print("Parents: " + ", ".join(product_extracted_version.child_of))


@product_app.command("delete")
def product_delete(id: str):
    """
    Delete a product by its ID.
    """
    global CLIENT
    return sc.product.delete(client=CLIENT, id=id, console=CONSOLE)


@product_app.command("search")
def product_search(text: str):
    """
    Search for products by name.
    """
    global CLIENT, CONSOLE

    response = sc.product.search(client=CLIENT, text=text, console=CONSOLE)

    table = helper.render_product_metadata_list(response)

    CONSOLE.print(table)


@product_app.command("cache")
def product_cache(id: str):
    """
    Cache a product by its ID.
    """
    global CLIENT, CACHE

    response = sc.product.cache(client=CLIENT, cache=CACHE, id=id, console=CONSOLE)

    CONSOLE.print(f"Cached product {id} including {len(response)} files")


@product_app.command("download")
def product_download(id: str, directory: str | None = None, slugify: bool = False):
    """
    Download a product to a location.
    """
    global CLIENT

    directory = Path(directory or ".")

    response = sc.product.download(
        client=CLIENT,
        id=id,
        directory=Path(directory),
        console=CONSOLE,
        slugify=slugify,
    )

    CONSOLE.print(f"Data cached to {response}")


@product_app.command("uncache")
def product_uncache(id: str):
    """
    Uncache a product by its ID.
    """
    global CACHE

    sc.product.uncache(client=CLIENT, cache=CACHE, id=id, console=CONSOLE)

    CONSOLE.print(f"Uncached product {id}")


@product_app.command("edit")
def product_edit(id: str):
    """
    Edit a product by its ID.
    """
    global CLIENT

    edit_product(client=CLIENT, id=id)


@product_app.command("add-reader")
def product_add_reader(id: str, group: str):
    """
    Add a reader (by group name) to a product
    """

    global CLIENT
    updated_id = sc.product.product_add_reader(
        client=CLIENT, id=id, group=group, console=CONSOLE
    )
    CONSOLE.print(f"Added {group} to {id} readers. New id is {updated_id}")


@product_app.command("remove-reader")
def product_remove_reader(id: str, group: str):
    """
    Remove a reader (by group name) from a product
    """

    global CLIENT
    updated_id = sc.product.product_remove_reader(
        client=CLIENT, id=id, group=group, console=CONSOLE
    )
    CONSOLE.print(f"Removed {group} from {id} readers. New id is {updated_id}")


@product_app.command("add-writer")
def product_add_writer(id: str, group: str):
    """
    Add a writer (by group name) to a product
    """

    global CLIENT
    updated_id = sc.product.product_add_writer(
        client=CLIENT, id=id, group=group, console=CONSOLE
    )
    CONSOLE.print(f"Added {group} to {id} writers. New id is {updated_id}")


@product_app.command("remove-writer")
def product_remove_writer(id: str, group: str):
    """
    Remove a writer (by group name) from a product
    """

    global CLIENT
    updated_id = sc.product.product_remove_writer(
        client=CLIENT, id=id, group=group, console=CONSOLE
    )
    CONSOLE.print(f"Removed {group} from {id} writers. New id is {updated_id}")


@product_app.command("add-child")
def product_add_child(parent: str, child: str):
    """
    Add a child relationship between two products.
    """
    global CLIENT

    sc.relationships.add_child(
        client=CLIENT, parent=parent, child=child, console=CONSOLE
    )
    CONSOLE.print(f"Added {child} as child of {parent}")


@product_app.command("remove-child")
def product_remove_child(parent: str, child: str):
    """
    Remove a child relationship between two products.
    """
    global CLIENT

    sc.relationships.remove_child(
        client=CLIENT, parent=parent, child=child, console=CONSOLE
    )
    CONSOLE.print(f"Removed {child} as child of {parent}")


@collection_app.command("read")
def collection_read(id: str):
    """
    Read the information of a collection by its name.
    """
    global CLIENT, CONSOLE

    collection = sc.collections.read(client=CLIENT, id=id, console=CONSOLE)

    table = helper.render_product_metadata_list(collection.products)

    CONSOLE.print(collection.name + "\n", style="bold underline color(3)")
    CONSOLE.print(rich.markdown.Markdown(collection.description.strip("\n")))
    CONSOLE.print("\n")
    CONSOLE.print(table)


@collection_app.command("search")
def collection_search(name: str):
    """
    Search for collections by name.
    """
    global CLIENT, CONSOLE

    collections = sc.collections.search(client=CLIENT, name=name, console=CONSOLE)

    table = helper.render_collection_metadata_list(collections)

    CONSOLE.print(table)


@collection_app.command("cache")
def collection_cache(id: str):
    """
    Cache a collection by its ID.
    """
    global CLIENT, CACHE, CONSOLE

    response = sc.collections.cache(client=CLIENT, cache=CACHE, id=id, console=CONSOLE)

    CONSOLE.print(f"Cached collection {id} including {len(response)} files")


@collection_app.command("download")
def collection_download(id: str, directory: str | None = None, slugify: bool = False):
    """
    Download a collection by its ID.
    """
    global CLIENT, CONSOLE

    directory = Path(directory or ".")

    response = sc.collections.download(
        client=CLIENT, id=id, directory=directory, console=CONSOLE, slugify=slugify
    )

    CONSOLE.print(f"Downloaded collection to {response}")


@collection_app.command("uncache")
def collection_uncache(id: str):
    """
    Uncache a collection by its ID.
    """
    global CACHE

    sc.collections.uncache(client=CLIENT, cache=CACHE, id=id, console=CONSOLE)

    CONSOLE.print(f"Uncached collection {id}")


@collection_app.command("delete")
def collection_delete(id: str):
    """
    Delete a collection by its ID.
    """
    global CLIENT

    sc.collections.delete(client=CLIENT, id=id, console=CONSOLE)

    CONSOLE.print(f"Deleted collection {id}")


@collection_app.command("add-product")
def collection_add_product(id: str, product_id: str):
    """
    Add a product to a collection.
    """
    global CLIENT
    updated_id = sc.collections.add(
        client=CLIENT, id=id, product=product_id, console=CONSOLE
    )
    CONSOLE.print(f"Added {product_id} to {id} collection. New ID is {updated_id}")


@collection_app.command("remove-product")
def collection_remove_product(id: str, product_id: str):
    """
    Remove a product from a collection.
    """
    global CLIENT
    updated_id = sc.collections.remove(
        client=CLIENT, id=id, product=product_id, console=CONSOLE
    )
    CONSOLE.print(f"Removed {product_id} from {id} collection. New ID is {updated_id}")


@collection_app.command("add-collection")
def collection_add_child(parent_id: str, child_id: str):
    """
    Add a sub-collection to a collection.
    """
    global CLIENT
    sc.relationships.add_child_collection(
        client=CLIENT, parent=parent_id, child=child_id, console=CONSOLE
    )
    CONSOLE.print(f"Added {child_id} to {parent_id} sub-collection")


@collection_app.command("remove-collection")
def collection_remove_child(parent_id: str, child_id: str):
    """
    Remove a sub-collection from a collection.
    """
    global CLIENT
    sc.relationships.remove_child_collection(
        client=CLIENT, parent=parent_id, child=child_id, console=CONSOLE
    )
    CONSOLE.print(f"Removed {child_id} from {parent_id} sub-collection")


@collection_app.command("add-reader")
def collection_add_reader(id: str, group: str):
    """
    Add a reader (by group name) to a collection.
    """
    global CLIENT
    updated_id = sc.collections.add_reader(
        client=CLIENT, id=id, group=group, console=CONSOLE
    )
    CONSOLE.print(f"Added {group} to readers. Collection ID is {updated_id}")


@collection_app.command("remove-reader")
def collection_remove_reader(id: str, group: str):
    """
    Remove a reader (by group name) from a collection.
    """
    global CLIENT
    updated_id = sc.collections.remove_reader(
        client=CLIENT, id=id, group=group, console=CONSOLE
    )
    CONSOLE.print(f"Removed {group} from readers. Collection ID is {updated_id}")


@collection_app.command("add-writer")
def collection_add_writer(id: str, group: str):
    """
    Add a writer (by group name) to a collection.
    """
    global CLIENT
    updated_id = sc.collections.add_writer(
        client=CLIENT, id=id, group=group, console=CONSOLE
    )
    CONSOLE.print(f"Added {group} to writers. Collection ID is {updated_id}")


@collection_app.command("remove-writer")
def collection_remove_writer(id: str, group: str):
    """
    Remove a writer (by group name) from a collection.
    """
    global CLIENT
    updated_id = sc.collections.remove_writer(
        client=CLIENT, id=id, group=group, console=CONSOLE
    )
    CONSOLE.print(f"Removed {group} from writers. Collection ID is {updated_id}")


@cache_app.command("clear")
def cache_clear(uuid: str):
    """
    Clear the cache of a single file, labelled by its UUID (note that product IDs don't work here).
    """
    global CACHE

    for cache in CACHE.caches:
        if cache.writeable:
            sc.caching.clear_single(cache=cache, id=id, console=CONSOLE)
            CONSOLE.print(f"Cleared cache {cache.path} of {id}")


@cache_app.command("clear-all")
def cache_clear_all():
    """
    Clear all caches of all files.
    """
    global CACHE

    for cache in CACHE.caches:
        if cache.writeable:
            sc.caching.clear_all(cache=cache)
            CONSOLE.print(f"Cleared cache {cache.path}")


@dev_app.command("serve")
def dev_serve(
    port: Annotated[int, typer.Option(help="Port to run the server on")] = 8000,
):
    """
    Run a hippo development server, without spinning up dependencies (i.e. you
    will need to have mongo and minio running with appropriate environments set).
    """

    import uvicorn

    from hipposerve.api.app import app

    config = uvicorn.Config(
        app,
        port=port,
        reload=True,
        reload_dirs=["hipposerve/web", "hipposerve/web/templates"],
    )
    server = uvicorn.Server(config)

    try:
        server.run()
    except KeyboardInterrupt:
        exit(0)


@dev_app.command("run")
def dev_run(
    with_soauth: Annotated[
        bool,
        typer.Option(
            help="Run with SOAuth integration? If not, all requests are mock-authenticated"
        ),
    ] = False,
    port: Annotated[int, typer.Option(help="Port to run the server on")] = 8000,
):
    """
    Run a hippo development server, spinning up dependencies (i.e. we create
    temporary servers and set defaults for all variables that we can).
    """

    import os

    import uvicorn
    from testcontainers.minio import MinioContainer
    from testcontainers.mongodb import MongoDbContainer

    database_kwargs = {
        "username": "root",
        "password": "password",
        "port": 27017,
        "dbname": "hippo_test",
    }

    storage_kwargs = {}

    with MongoDbContainer(**database_kwargs) as database_container:
        with MinioContainer(**storage_kwargs) as storage_container:
            storage_config = storage_container.get_config()
            database_uri = database_container.get_connection_url()

            settings = {
                "mongo_uri": database_uri,
                "minio_url": storage_config["endpoint"],
                "minio_access": storage_config["access_key"],
                "minio_secret": storage_config["secret_key"],
                "title": "Test hippo",
                "description": "Test hippo Description",
                "debug": "yes",
                "add_cors": "yes",
                "web": "yes",
                "auth_system": "soauth" if with_soauth else "None",
            }

            os.environ.update(settings)

            # HIPPO must be imported _after_ updating settings.
            from hipposerve.api.app import app

            config = uvicorn.Config(
                app,
                port=port,
                reload=True,
                reload_dirs=["hipposerve/web", "hipposerve/web/templates"],
            )
            server = uvicorn.Server(config)

            try:
                server.run()
            except KeyboardInterrupt:
                exit(0)

    return


def main():
    settings = ClientSettings()

    global CLIENT, APP, CACHE, CONSOLE

    CLIENT = settings.client
    CACHE = settings.cache
    CONSOLE = rich.console.Console(quiet=not settings.verbose)

    APP()



"""
Metadata specification for the Simons Observatory Post Office.

This package provides pydantic models describing all of the metadata
types that the Post Office can handle, and potentially some
utility functions for automatically filling out the structures with
data.

Each metadata type can provide utility functions as a class method.
Each metadata type must inherit from the BaseMetadata type defined here.
Each metadata type must have a unique `metadata_type` field which is a
string used as a discriminator.
Each metadata type must be added to the below ALL_METADATA and
ALL_METADATA_TYPE lists.
"""

from typing import Annotated, Union

from pydantic import Field

from hippometa.archive import ArchiveMetadata
from hippometa.beam import BeamMetadata
from hippometa.cameras import CameraMetadata
from hippometa.catalog import CatalogMetadata
from hippometa.mapset import MapSet
from hippometa.numeric import NumericMetadata
from hippometa.simple import SimpleMetadata

ALL_METADATA_TYPE = Annotated[
    Union[
        ArchiveMetadata,
        BeamMetadata,
        MapSet,
        CatalogMetadata,
        NumericMetadata,
        SimpleMetadata,
        CameraMetadata,
        None,
    ],
    Field(discriminator="metadata_type"),
]

ALL_METADATA = {
    x.model_fields["metadata_type"].default: x
    for x in [
        ArchiveMetadata,
        BeamMetadata,
        MapSet,
        CatalogMetadata,
        NumericMetadata,
        SimpleMetadata,
        CameraMetadata,
    ]
}



"""
The product service layer.
"""

import datetime
import re
from typing import Any, Literal

from beanie import Link, PydanticObjectId, WriteRules
from beanie.operators import Text
from bson.errors import InvalidId
from loguru import logger
from pydantic import BaseModel
from pydantic_core import ValidationError

from hippometa import ALL_METADATA_TYPE, SimpleMetadata
from hipposerve.database import (
    Collection,
    CollectionPolicy,
    File,
    Product,
    ProductMetadata,
)
from hipposerve.service import storage as storage_service
from hipposerve.service import utils, versioning
from hipposerve.service.acl import check_user_access
from hipposerve.storage import Storage

INITIAL_VERSION = "1.0.0"
LINK_POLICY = {
    "fetch_links": True,
    "nesting_depth": 2,
}


class ProductExists(Exception):
    pass


class ProductNotFound(Exception):
    pass


class PreUploadFile(BaseModel):
    name: str
    size: int
    checksum: str
    description: str | None = None


class PostUploadFile(BaseModel):
    uuid: str
    name: str
    size: int
    slug: str
    checksum: str
    url: str | None
    description: str | None
    object_name: str | None
    available: bool


async def presign_uploads(
    sources: dict[str, PreUploadFile],
    storage: Storage,
    user_name: str,
    mutlipart_size: int = 50 * 1024 * 1024,
) -> tuple[dict[str, list[str]], dict[str, File]]:
    """
    Note that upload pre-signing does _not_ validate sources' slugs as it does not
    know about the metadata type.
    """
    presigned = {}
    pre_upload_sources = {}

    for slug, source in sources.items():
        pre_upload_source, presigned_urls = await storage_service.create(
            name=source.name,
            description=source.description,
            slug=slug,
            uploader=user_name,
            size=source.size,
            checksum=source.checksum,
            storage=storage,
        )

        presigned[source.name] = presigned_urls
        pre_upload_sources[slug] = pre_upload_source

    return presigned, pre_upload_sources


async def exists(name: str) -> bool:
    """
    Check whether a product exists with this name.
    Product names are unique but this cannot be enforced at the
    database level due to versioning.
    """
    return (await Product.find(Product.name == name).count()) > 0


def validate_slugs(sources: dict[str, PreUploadFile], metadata: ALL_METADATA_TYPE):
    """
    Ensures that the slugs available in sources are valid according
    to the metadata type. If they are not, raises a pydantic ValidationError.
    """

    errors = []

    for k in sources.keys():
        if k not in metadata.valid_slugs:
            errors.append(
                ValueError(
                    f"Slug '{k}' not part of acceptable slugs: {metadata.valid_slugs} for this metadata type"
                )
            )

    if errors:
        raise ExceptionGroup("Invalid slugs", errors)


async def create(
    name: str,
    description: str,
    metadata: ALL_METADATA_TYPE,
    sources: dict[str, PreUploadFile],
    user_name: str,
    storage: Storage,
    product_readers: list[str] | None = None,
    product_writers: list[str] | None = None,
    mutlipart_size: int = 50 * 1024 * 1024,
) -> tuple[Product, dict[str, list[str]]]:
    if metadata is None:
        metadata = SimpleMetadata()

    validate_slugs(sources=sources, metadata=metadata)

    presigned, pre_upload_sources = await presign_uploads(
        sources=sources,
        storage=storage,
        user_name=user_name,
        mutlipart_size=mutlipart_size,
    )

    current_utc_time = datetime.datetime.now(datetime.timezone.utc)

    product = Product(
        name=name,
        description=description,
        metadata=metadata,
        uploaded=current_utc_time,
        updated=current_utc_time,
        current=True,
        version=INITIAL_VERSION,
        sources=pre_upload_sources,
        owner=user_name,
        replaces=None,
        child_of=[],
        parent_of=[],
        collections=[],
        readers=set(product_readers or []) | {user_name},
        writers=set(product_writers or []) | {user_name},
    )

    await product.create()

    return product, presigned


async def read_by_name(
    name: str, version: str | None, groups: list[str], scopes: set[str]
) -> Product:
    """
    If version is None, we grab the latest version of a product.
    """

    if version is None:
        potential = await Product.find_one(
            Product.name == name,
            Product.current == True,  # noqa: E712
            **LINK_POLICY,
        )
    else:
        potential = await Product.find_one(
            Product.name == name,
            Product.version == version,
            **LINK_POLICY,
        )

    if potential is None:
        raise ProductNotFound

    assert check_user_access(
        user_groups=groups,
        document_groups=potential.readers + potential.writers,
        scopes=scopes,
    )

    return potential


async def read_by_id(
    id: PydanticObjectId, groups: list[str], scopes: set[str]
) -> Product:
    try:
        potential = await Product.get(document_id=id, **LINK_POLICY)
    except (InvalidId, ValidationError, ValueError):
        raise ProductNotFound

    if potential is None:
        raise ProductNotFound

    assert check_user_access(
        user_groups=groups,
        document_groups=potential.readers + potential.writers,
        scopes=scopes,
    )

    return potential


async def search_by_name(
    name: str, groups: list[str], scopes: set[str], fetch_links: bool = True
) -> list[Product]:
    """
    Search for products by name using the text index.
    """

    if "hippo:admin" in scopes:
        access_query = {}
    else:
        access_query = {
            "$or": [{"readers": {"$in": groups}}, {"writers": {"$in": groups}}]
        }

    results = (
        await Product.find(
            access_query,
            Text(name),
            Product.current == True,  # noqa: E712
            fetch_links=fetch_links,
        )
        .sort([("score", {"$meta": "textScore"})])
        .to_list()
    )

    return results


async def search_by_metadata(
    metadata_filters: dict[str, Any],
    groups: list[str],
    scopes: set[str],
    fetch_links: bool = True,
) -> list[Product]:
    """
    Search for products by metadata.
    """

    if "hippo:admin" in scopes:
        access_query = {}
    else:
        access_query = {
            "$or": [{"readers": {"$in": groups}}, {"writers": {"$in": groups}}]
        }

    # Construct the query by embedding metadata field names and values
    query = {f"metadata.{key}": value for key, value in metadata_filters.items()}

    # Execute the query
    results = await Product.find(access_query, query, fetch_links=fetch_links).to_list()

    return results


async def search_by_owner(
    owner: str,
    groups: list[str],
    scopes: set[str],
    fetch_links: bool = True,
) -> list[Collection]:
    """
    Search for products by owner; owner search is case insensitive
    but must otherwise be an exact match
    """

    owner_regex = {"$regex": f"^{re.escape(owner)}$", "$options": "i"}

    if "hippo:admin" in scopes:
        access_query = {}
    else:
        access_query = {
            "$or": [{"readers": {"$in": groups}}, {"writers": {"$in": groups}}]
        }

    results = await Product.find(
        {**access_query, "owner": owner_regex},
        Product.current == True,  # noqa: E712
        fetch_links=fetch_links,
    ).to_list()

    return results


async def walk_history(product: Product) -> dict[str, ProductMetadata]:
    """
    Walk the history of the product from this point backwards.
    It is recommended that you walk to the current version first,
    to get the whole history.
    """
    versions = {}

    while product.replaces is not None:
        versions[product.version] = await product.to_metadata()
        product = product.replaces

        if isinstance(product, Link):
            product = await product.fetch(fetch_links=True)

    versions[product.version] = await product.to_metadata()

    return versions


async def walk_to_current(
    product: Product, groups: list[str], scopes: set[str]
) -> Product:
    """
    Walk the list of products until you get to the one
    marked 'current'.
    """

    # Re-read the product from the database, in case it
    # is stale!
    product = await read_by_id(id=product.id, groups=groups, scopes=scopes)

    while not product.current:
        product = await Product.find_one(
            Product.replaces.id == product.id, **LINK_POLICY
        )

        if product is None:  # pragma: no cover
            raise RuntimeError

    return product


async def complete(
    product: Product,
    storage: Storage,
    headers: dict[str, list[dict[str, str]]],
    sizes: dict[str, list[int]],
) -> bool:
    for file in product.sources.values():
        if file.multipart_closed:
            continue

        await storage_service.complete(
            file=file,
            storage=storage,
            response_headers=headers[file.name],
            sizes=sizes[file.name],
        )

        file.multipart_closed = True
        await file.save()

    return True


async def confirm(product: Product, storage: Storage) -> bool:
    for slug, file in product.sources.items():
        if not await storage_service.confirm(file=file, storage=storage):
            logger.debug(f"File {file.name} ({slug}) not confirmed")
            return False

    return True


async def read_files(product: Product, storage: Storage) -> dict[str, PostUploadFile]:
    return {
        slug: PostUploadFile(
            uuid=x.uuid,
            name=x.name,
            size=x.size,
            slug=x.slug,
            checksum=x.checksum,
            description=x.description,
            url=storage.get(
                name=x.name, uploader=x.uploader, uuid=x.uuid, bucket=x.bucket
            )
            if x.available
            else None,
            object_name=storage.object_name(
                filename=x.name, uploader=x.uploader, uuid=x.uuid
            )
            if x.available
            else None,
            available=x.available,
        )
        for slug, x in product.sources.items()
    }


async def read_most_recent(
    groups: list[str],
    scopes: set[str],
    fetch_links: bool = False,
    maximum: int = 16,
    current_only: bool = True,
) -> list[Product]:
    if "hippo:admin" in scopes:
        access_query = {}
    else:
        access_query = {
            "$or": [{"readers": {"$in": groups}}, {"writers": {"$in": groups}}]
        }

    if current_only:
        query = {"$and": [{"current": True}, access_query]}
    else:
        query = access_query

    found = Product.find(
        query,
        fetch_links=fetch_links,
    )
    return await found.sort(-Product.updated).to_list(maximum)


async def read_pinned(
    groups: list[str],
    scopes: set[str],
    fetch_links: bool = False,
    maximum: int = 16,
    current_only: bool = True,
) -> list[Product]:
    if "hippo:admin" in scopes:
        access_query = {"pinned": True}
    else:
        access_query = {
            "$and": [
                {"pinned": True},
                {"$or": [{"readers": {"$in": groups}}, {"writers": {"$in": groups}}]},
            ]
        }

    if current_only:
        access_query = {"$and": [access_query, {"current": True}]}

    found = Product.find(
        access_query,
        fetch_links=fetch_links,
    )

    return await found.sort(-Product.updated).to_list(maximum)


async def metadata_diff(
    product: Product,
    name: str | None,
    description: str | None,
    metadata: ALL_METADATA_TYPE | None,
    level: versioning.VersionRevision,
):
    """
    Calculate the diff between the current product and the new metadata.
    This is used to determine what has changed and what needs to be updated.
    """

    if not product.current:
        raise versioning.VersioningError(
            "Attempting to update a non-current product. You must always "
            "make changes to the head of the list"
        )

    changes = {}

    if name is not None and name != product.name:
        changes["name"] = [product.name, name]

    if description is not None and description != product.description:
        changes["description"] = [product.description, description]

    if metadata is not None:
        metadata_diff = {}
        for key, value in metadata.model_dump().items():
            original_value = getattr(product.metadata, key, None)

            if value != original_value:
                metadata_diff[key] = [original_value, value]

        if metadata_diff:
            changes["metadata"] = metadata_diff

    changes["version"] = [
        product.version,
        versioning.revise_version(product.version, level=level),
    ]

    return changes


async def update_metadata(
    product: Product,
    name: str | None,
    description: str | None,
    metadata: ALL_METADATA_TYPE | None,
    level: versioning.VersionRevision,
) -> Product:
    """
    Update only the metadata of a product. Note that you can call this with all
    the variables as `None` and this will simply rev the version. That should
    be used when only updating sources.
    """

    if not product.current:
        raise versioning.VersioningError(
            "Attempting to update a non-current product. You must always "
            "make changes to the head of the list"
        )

    # We don't actually 'update' the database; we actually create a new
    # product and link it in.

    new = Product(
        name=product.name if name is None else name,
        description=product.description if description is None else description,
        metadata=product.metadata if metadata is None else metadata,
        uploaded=product.uploaded,
        owner=product.owner,
        updated=utils.current_utc_time(),
        current=True,
        version=versioning.revise_version(product.version, level=level),
        sources=product.sources,
        replaces=product,
        # This product is a child of nothing util it is told it is. Child
        # relationships only stick around for a single version.
        child_of=[],
        collections=[
            c
            for c, p in zip(product.collections, product.collection_policies)
            if p
            in [CollectionPolicy.ALL, CollectionPolicy.NEW, CollectionPolicy.CURRENT]
        ],
        collection_policies=[
            p
            for p in product.collection_policies
            if p
            in [CollectionPolicy.ALL, CollectionPolicy.NEW, CollectionPolicy.CURRENT]
        ],
        readers=product.readers,
        writers=product.writers,
    )

    # Need to perform a small number of modifications on the original
    # product.
    product.current = False
    product.collections = [
        c
        for c, p in zip(product.collections, product.collection_policies)
        if p in [CollectionPolicy.ALL, CollectionPolicy.NEW, CollectionPolicy.FIXED]
    ]
    product.collection_policies = [
        p
        for p in product.collection_policies
        if p in [CollectionPolicy.ALL, CollectionPolicy.NEW, CollectionPolicy.FIXED]
    ]

    await new.save(link_rule=WriteRules.WRITE)
    await product.save(link_rule=WriteRules.WRITE)

    return new


async def update_sources(
    product: Product,
    new: dict[str, PreUploadFile],
    replace: dict[str, PreUploadFile],
    drop: list[str],
    storage: Storage,
) -> tuple[Product, dict[str, str]]:
    """
    Change the sources associated with a product. New and replace provide
    pre-signed URLs for file uploads (and you should call confirm after),
    and drop is a list of names of files to remove.
    """

    existing_slugs = set(product.sources.keys())
    new_slugs = set(new.keys())
    replace_slugs = set(replace.keys())
    drop_slugs = set(drop)

    if len(existing_slugs & new_slugs) != 0:
        raise FileExistsError

    if len(existing_slugs & replace_slugs) != len(replace_slugs):
        raise FileNotFoundError

    if len(existing_slugs & drop_slugs) != len(drop_slugs):
        raise FileNotFoundError

    # Check our slugs are valid for this metadata type
    validate_slugs(sources={**new, **replace}, metadata=product.metadata)

    presigned_new, pre_upload_sources_new = await presign_uploads(
        sources=new,
        storage=storage,
        user_name=product.owner,
    )

    presigned_replace, pre_upload_sources_replace = await presign_uploads(
        sources=replace, storage=storage, user_name=product.owner
    )

    presigned = {**presigned_new, **presigned_replace}
    pre_upload_sources = {**pre_upload_sources_new, **pre_upload_sources_replace}

    # Filter out the sources we're keeping
    keep_slugs = existing_slugs ^ replace_slugs ^ drop_slugs
    keep_sources = {x: y for x, y in product.sources.items() if x in keep_slugs}

    # Final check -
    expected_number_of_sources = len(product.sources) - len(drop) + len(new)

    if (
        not (len(keep_sources) + len(pre_upload_sources)) == expected_number_of_sources
    ):  # pragma: no cover
        raise RuntimeError

    product.sources = {**pre_upload_sources, **keep_sources}

    await product.save(link_rule=WriteRules.WRITE)

    return product, presigned


async def update(
    product: Product,
    access_groups: list[str],
    scopes: set[str],
    storage: Storage,
    level: versioning.VersionRevision,
    name: str | None = None,
    description: str | None = None,
    metadata: ALL_METADATA_TYPE | None = None,
    new_sources: dict[str, PreUploadFile] | None = None,
    replace_sources: dict[str, PreUploadFile] | None = None,
    drop_sources: list[str] | None = None,
) -> tuple[Product, dict[str, str]]:
    """
    Rev the version of the product and update its contents.
    """
    assert check_user_access(
        user_groups=access_groups, document_groups=product.writers, scopes=scopes
    )

    new_product = await update_metadata(
        product=product,
        name=name,
        description=description,
        metadata=metadata,
        level=level,
    )

    new_sources = new_sources or {}
    replace_sources = replace_sources or {}
    drop_sources = drop_sources or []

    # For some reason:
    # a) Beanie won't allow fetch_all_links to be called on `new_product` because
    #    it has back-links, and it crashes.
    # b) Beanie won't allow fetching specific links on that object. fetch_link()
    #    or .fetch() fails silently.
    # So we just re-fetch our object from the database.

    new_product = await read_by_id(
        new_product.id, new_product.writers + new_product.readers, scopes=scopes
    )  # who should read this product?

    if any([len(new_sources) > 0, len(replace_sources) > 0, len(drop_sources) > 0]):
        try:
            new_product, presigned = await update_sources(
                product=new_product,
                new=new_sources,
                replace=replace_sources,
                drop=drop_sources,
                storage=storage,
            )
        except Exception as e:
            # Need to roll-back our new product. Full transactions when?
            await delete_one(
                new_product,
                [new_product.owner],
                storage=storage,
                data=False,
                scopes=scopes,
            )
            raise e
    else:
        presigned = {}

    return new_product, presigned


async def pin(
    product: Product,
    access_groups: list[str],
    scopes: set[str],
):
    """
    Pin a product in the database.
    """

    assert check_user_access(
        user_groups=access_groups, document_groups=product.writers, scopes=scopes
    )

    product.pinned = True
    await product.save()

    return


async def unpin(
    product: Product,
    access_groups: list[str],
    scopes: set[str],
):
    """
    Unpin a product in the database.
    """

    assert check_user_access(
        user_groups=access_groups, document_groups=product.writers, scopes=scopes
    )

    product.pinned = False
    await product.save()

    return


async def delete_one(
    product: Product,
    access_groups: list[str],
    scopes: set[str],
    storage: Storage,
    data: bool = False,
):
    """
    Delete a single product (i.e. leaving the history intact on either side.)
    """

    # Deal with parent/child replacement relationship
    assert check_user_access(
        user_groups=access_groups, document_groups=product.writers, scopes=scopes
    )
    if product.current:
        replaced_by = None
        replaces = product.replaces
        if replaces is not None:
            replaces.current = True
    else:
        replaced_by = await Product.find_one(Product.replaces.id == product.id)
        replaces = product.replaces
        replaced_by.replaces = replaces

    # Deal with potentially newly ingested files

    if replaces is None:
        past_source_uuids = set()
    else:
        past_source_uuids = {x.uuid for x in replaces.sources.values()}

    current_source_uuids = {x.uuid for x in product.sources.values()}

    if replaced_by is None:
        future_source_uuids = set()
    else:
        future_source_uuids = {x.uuid for x in replaced_by.sources.values()}

    net_source_uuids = current_source_uuids ^ past_source_uuids ^ future_source_uuids

    # Deletion of only uniquely created (in this specific version of the product)
    # data products.

    if data:
        for file in product.sources.values():
            if file.uuid in net_source_uuids:
                await storage_service.delete(file=file, storage=storage)

    if replaced_by is not None:
        await replaced_by.save()

    if replaces is not None:
        await replaces.save()

    await product.delete()

    return


async def delete_tree(
    product: Product,
    access_groups: list[str],
    scopes: set[str],
    storage: Storage,
    data: bool = False,
):
    """
    Delete an entire tree of products, from this one down. You must provide
    a current product.
    """
    assert check_user_access(
        user_groups=access_groups, document_groups=product.writers, scopes=scopes
    )
    if not product.current:
        raise versioning.VersioningError(
            "Attempting to delete the tree starting from a non-current product"
        )

    uuids_to_delete = set()
    files_to_delete = []
    products_to_delete = []

    while product is not None:
        products_to_delete.append(product)

        for source in product.sources.values():
            if source.uuid not in uuids_to_delete:
                uuids_to_delete.update(source.uuid)
                files_to_delete.append(source)

        if isinstance(product.replaces, Link):
            product = await product.replaces.fetch()
        else:
            product = product.replaces

    if data:
        for source in files_to_delete:
            await storage_service.delete(file=source, storage=storage)

    for product in products_to_delete:
        await product.delete()

    return


async def add_relationship(
    source: Product,
    destination: Product,
    access_groups: list[str],
    scopes: set[str],
    type: Literal["child"],
):
    if type == "child":
        source.child_of = source.child_of + [destination]
    assert check_user_access(
        user_groups=access_groups, document_groups=source.writers, scopes=scopes
    )
    await source.save()

    return


async def remove_relationship(
    source: Product,
    destination: Product,
    access_groups: list[str],
    scopes: set[str],
    type: Literal["child"],
):
    if type == "child":
        source.child_of = [c for c in source.child_of if c.id != destination.id]
    assert check_user_access(
        user_groups=access_groups, document_groups=source.writers, scopes=scopes
    )
    await source.save()

    return


async def add_collection(
    product: Product, access_groups: list[str], scopes: set[str], collection: Collection
):
    product.collections = product.collections + [collection]
    assert check_user_access(
        user_groups=access_groups, document_groups=product.writers, scopes=scopes
    )
    await product.save()

    return


async def remove_collection(
    product: Product, access_groups: list[str], scopes: set[str], collection: Collection
):
    product.collections = [c for c in product.collections if c.id != collection.id]
    assert check_user_access(
        user_groups=access_groups, document_groups=product.writers, scopes=scopes
    )
    await product.save()

    return product



from pathlib import Path

from httpx import Client
from rich.console import Console

from henry.source import LocalSource
from hippoclient.caching import MultiCache
from hippoclient.core import Client as AuthenticatedClient
from hippoclient.core import ClientSettings
from hippoclient.product import download as download_product
from hippometa import ALL_METADATA_TYPE

from .collection import CollectionInstance, LocalCollection, RemoteCollection
from .product import LocalProduct, ProductInstance, RemoteProduct


class Henry:
    settings: ClientSettings
    client: Client
    console: Console
    cache: MultiCache
    slugify: bool

    def __init__(
        self,
        *,
        settings: ClientSettings | None = None,
        console: Console | None = None,
        slugify: bool = False,
    ):
        self.settings = settings or ClientSettings()
        self.console = console or Console(quiet=(not self.settings.verbose))
        self.client = AuthenticatedClient(
            host=self.settings.host, token_tag=self.settings.token_tag
        )
        self.cache = self.settings.cache
        self.readers = self.settings.default_readers
        self.writers = self.settings.default_writers
        self.slugify = slugify

        return

    def new_product(
        self,
        name: str,
        description: str,
        metadata: ALL_METADATA_TYPE,
        sources: dict[str, LocalSource] | None = None,
    ) -> LocalProduct:
        return LocalProduct(
            name=name, description=description, metadata=metadata, sources=sources or {}
        )

    def new_collection(
        self,
        name: str,
        description: str,
        products: list[ProductInstance] | None = None,
        collections: list[CollectionInstance] | None = None,
    ) -> LocalCollection:
        return LocalCollection(
            name=name,
            description=description,
            products=products or [],
            collections=collections or [],
        )

    def push(
        self, item: LocalProduct | LocalCollection, skip_preflight: bool = False
    ) -> str:
        return item._upload(
            client=self.client,
            console=self.console,
            skip_preflight=skip_preflight,
            readers=self.readers,
            writers=self.writers,
        )

    def pull_product(
        self, product_id: str, realize_sources: bool = True
    ) -> RemoteProduct:
        return RemoteProduct.pull(
            product_id=product_id,
            client=self.client,
            cache=self.cache,
            console=self.console,
            realize_sources=realize_sources,
        )

    def read_product(
        self, directory: str | Path, allow_incomplete: bool = True
    ) -> RemoteProduct:
        return RemoteProduct.read(
            directory=directory, allow_incomplete=allow_incomplete
        )

    def pull_collection(
        self,
        collection_id: str,
        realize_sources: bool = True,
        pull_children: bool = True,
    ) -> CollectionInstance:
        return RemoteCollection.pull(
            collection_id=collection_id,
            client=self.client,
            cache=self.cache,
            console=self.console,
            pull_children=pull_children,
            realize_sources=realize_sources,
        )

    def read_collection(
        self, directory: str | Path, allow_incomplete: bool = True
    ) -> RemoteCollection:
        return RemoteCollection.read(
            directory=directory, allow_incomplete=allow_incomplete
        )

    def download_product(
        self, product_id: str, directory: str | Path, slugs: list[str] | None = None
    ) -> Path:
        return download_product(
            client=self.client,
            id=product_id,
            directory=Path(directory),
            console=self.console,
            slugs=slugs,
            slugify=self.slugify,
        )



"""
Methods for adding and removing child relationships between products.
"""

from httpx import Client
from rich.console import Console


def add_child(
    client: Client, parent: str, child: str, console: Console | None = None
) -> bool:
    """
    Add a child relationship between two products.

    Arguments
    ---------
    client : Client
        The client to use for interacting with the hippo API.
    parent : str
        The ID of the parent product.
    child : str
        The ID of the child product.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    bool
        True if the relationship was added successfully.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.put(f"/relationships/product/{child}/child_of/{parent}")

    response.raise_for_status()

    if console:
        console.print(
            f"Successfully added child relationship between {parent} and {child}.",
            style="bold green",
        )

    return True


def remove_child(
    client: Client, parent: str, child: str, console: Console | None = None
) -> bool:
    """
    Remove a child relationship between two products.

    Arguments
    ---------
    client : Client
        The client to use for interacting with the hippo API.
    parent : str
        The ID of the parent product.
    child : str
        The ID of the child product.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    bool
        True if the relationship was removed successfully.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.delete(f"/relationships/product/{child}/child_of/{parent}")

    response.raise_for_status()

    if console:
        console.print(
            f"Successfully removed child relationship between {parent} and {child}.",
            style="bold green",
        )

    return True


def add_child_collection(
    client: Client, parent: str, child: str, console: Console | None = None
) -> bool:
    """
    Add a child relationship between a collection and another collection.

    Arguments
    ---------
    client : Client
        The client to use for interacting with the hippo API.
    parent : str
        The ID of the parent collection.
    child : str
        The ID of the child collection.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    bool
        True if the relationship was added successfully.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.put(f"/relationships/collection/{parent}/child_of/{child}")

    response.raise_for_status()

    if console:
        console.print(
            f"Successfully added child relationship between {parent} and {child}.",
            style="bold green",
        )

    return True


def remove_child_collection(
    client: Client, parent: str, child: str, console: Console | None = None
) -> bool:
    """
    Remove a child relationship between a collection and another collection.

    Arguments
    ---------
    client : Client
        The client to use for interacting with the hippo API.
    parent : str
        The ID of the parent collection.
    child : str
        The ID of the child collection.
    console : Console, optional
        The rich console to print to.

    Returns
    -------
    bool
        True if the relationship was removed successfully.

    Raises
    ------
    httpx.HTTPStatusError
        If a request to the API fails
    """

    response = client.delete(f"/relationships/collection/{child}/child_of/{parent}")

    response.raise_for_status()

    if console:
        console.print(
            f"Successfully removed child relationship between {parent} and {child}.",
            style="bold green",
        )

    return True



"""
Tools for version management.
"""

from enum import Enum


class VersioningError(Exception):
    pass


class VersionRevision(Enum):
    MAJOR = 0
    MINOR = 1
    PATCH = 2


def revise_version(current: str, level: VersionRevision) -> str:
    split = [int(x) for x in current.split(".")]

    split[level.value] = split[level.value] + 1

    for index in range(level.value + 1, len(split)):
        split[index] = 0

    return ".".join([f"{x}" for x in split])



"""
Authentication layer for the Web UI. Provides two core dependencies:

- `LoggedInUser` - a dependency that requires a user to be logged in. If the user is not logged in, a 401 Unauthorized response is returned.
- `PotentialLoggedInUser` - a dependency that returns a user if they are logged in, or `None` if they are not.
"""

from typing import Annotated
from urllib.parse import urlparse, urlunparse

from fastapi import APIRouter, Depends, HTTPException, Request, status
from fastapi.responses import RedirectResponse
from loguru import logger

from hipposerve.service import collection, product
from hipposerve.service import users as users_service
from hipposerve.settings import SETTINGS

from .router import templates


def replace_path_component(url: str, old: str, new: str) -> str:
    """
    Replace a path component in a URL.

    Parameters
    ----------
    url : str
        The URL to modify.
    old : str
        The old path component.
    new : str
        The new path component.

    Returns
    -------
    str
        The modified URL.
    """

    parsed = urlparse(url)
    fixed = parsed._replace(path=parsed.path.replace(old, new))
    return urlunparse(fixed)


def generate_redirect_response(
    request: Request, old: str, new: str
) -> RedirectResponse:
    """
    Generate a redirect response with a path component replaced.

    Parameters
    ----------
    request : Request
        The request object.
    old : str
        The old path component.
    new : str
        The new path component.

    Returns
    -------
    RedirectResponse
        The redirection response (code 302)
    """
    url = str(request.url)
    new_url = replace_path_component(url, old, new)

    return RedirectResponse(url=new_url, status_code=302)


router = APIRouter()


class UnauthorizedException(HTTPException):
    def __init__(
        self, detail: str | None = "Authentication required to access this page."
    ):
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=detail,
        )


async def get_current_user(request: Request):
    """
    Returns the authenticated HIPPO user or raises UnauthorizedException.
    """
    user = request.user
    if not request.user.is_authenticated:
        raise UnauthorizedException()
    return user


async def get_potential_current_user(request: Request):
    try:
        return await get_current_user(request)
    except HTTPException:
        return None
    return None


LoggedInUser = Annotated[users_service.User, Depends(get_current_user)]
PotentialLoggedInUser = Annotated[
    users_service.User | None, Depends(get_potential_current_user)
]


@router.get("/user")
async def read_user(request: Request):
    username = request.user.display_name
    collections = (
        await collection.search_by_owner(
            username, request.user.groups, scopes=request.auth.scopes
        )
        or []
    )
    products = (
        await product.search_by_owner(
            username, request.user.groups, scopes=request.auth.scopes
        )
        or []
    )

    logger.info(
        "User page for {} has {} products and {} collections",
        username,
        len(products),
        len(collections),
    )

    return templates.TemplateResponse(
        "user.html",
        {
            "request": request,
            "products": products,
            "collections": collections,
            "web_root": SETTINGS.web_root,
        },
    )


@router.get("/login", name="login")
async def login(request: Request):
    return RedirectResponse(url=request.app.login_url, status_code=302)



"""
Facilities for creating and updating users.
"""

from beanie import PydanticObjectId

from hipposerve.database import User
from hipposerve.service.utils import current_utc_time


# TODO: Centralised exceptions?
class UserNotFound(Exception):
    pass


class AuthenticationError(Exception):
    pass


async def create(
    name: str,
    email: str | None,
) -> User:
    user = User(
        name=name,
        email=email,
        last_access_time=current_utc_time(),
    )
    await user.create()
    return user


async def touch_last_access_time(name: str) -> None:
    user = await User.find(User.name == name).first_or_none()

    if user is None:
        raise UserNotFound

    user.last_access_time = current_utc_time()
    await user.save()

    return


async def confirm_user(name: str) -> None:
    exists = await User.find(User.name == name).exists()
    if not exists:
        raise UserNotFound


async def read(name: str) -> User:
    result = await User.find(User.name == name, fetch_links=True).first_or_none()

    if result is None:
        raise UserNotFound

    return result


async def read_by_id(id: PydanticObjectId) -> User:
    result = await User.get(id, fetch_links=True)

    if result is None:
        raise UserNotFound

    return result


async def delete(name: str):
    user = await read(name=name)
    await user.delete()

    return



"""
Service drivers for interacting with the storage layer.
"""

import os
import uuid
from math import ceil

from asyncer import asyncify

from hipposerve.database import File
from hipposerve.storage import Storage


# TODO: Settings
def UUID():
    return str(uuid.uuid4())


GLOBAL_BUCKET_NAME = "global"


async def create(
    name: str,
    description: str | None,
    slug: str,
    uploader: str,
    size: int,
    checksum: str,
    storage: Storage,
    multipart_size: int = 50 * 1024 * 1024,
) -> tuple[File, str]:
    # All uploads are now multipart.
    multipart = True
    number_of_parts = int(ceil(size / multipart_size))
    uuid = UUID()

    upload_id, put = await asyncify(storage.put)(
        name=name,
        uploader=uploader,
        uuid=uuid,
        bucket=GLOBAL_BUCKET_NAME,
        size=size,
        batch=multipart_size,
    )

    file = File(
        # Strip any paths that were passed to us through
        # the layers, just in case.
        name=os.path.basename(name),
        description=description,
        slug=slug,
        uploader=uploader,
        uuid=uuid,
        bucket=GLOBAL_BUCKET_NAME,
        size=size,
        checksum=checksum,
        multipart=multipart,
        number_of_parts=number_of_parts,
        upload_id=upload_id,
        multipart_batch_size=multipart_size,
        multipart_closed=not multipart,
    )

    return file, put


async def complete(
    file: File,
    storage: Storage,
    response_headers: list[dict[str, str]],
    sizes: list[int],
):
    """
    Multipart uploads must come with a complete step, including responses from
    the storage server along the way.
    """

    await asyncify(storage.complete)(
        name=file.name,
        uploader=file.uploader,
        uuid=file.uuid,
        bucket=file.bucket,
        upload_id=file.upload_id,
        headers=response_headers,
        sizes=sizes,
    )

    return


async def confirm(file: File, storage: Storage) -> bool:
    return await asyncify(storage.confirm)(
        name=file.name, uploader=file.uploader, uuid=file.uuid, bucket=file.bucket
    )


async def read(
    file: File,
    storage: Storage,
) -> str:
    return await asyncify(storage.get)(
        name=file.name, uploader=file.uploader, uuid=file.uuid, bucket=file.bucket
    )


async def delete(
    file: File,
    storage: Storage,
):
    await asyncify(storage.delete)(
        name=file.name, uploader=file.uploader, uuid=file.uuid, bucket=file.bucket
    )
    return



"""
Service layer for collections.

Note that adding to and removing from collections is provided as part of
the product service.
"""

import re

from beanie import PydanticObjectId
from beanie.operators import Text
from fastapi import HTTPException

from hipposerve.database import Collection
from hipposerve.service.acl import check_user_access

LINK_POLICY = {
    "fetch_links": True,
    "nesting_depth": 1,
}


class CollectionNotFound(Exception):
    pass


async def create(
    name: str,
    user: str,
    description: str,
    collection_readers: list[str] | None = None,
    collection_writers: list[str] | None = None,
):
    collection = Collection(
        name=name,
        owner=user,
        description=description,
        readers=set(collection_readers or []) | {user},
        writers=set(collection_writers or []) | {user},
    )

    await collection.insert()

    return collection


async def read(id: PydanticObjectId, groups: list[str], scopes: set[str]):
    collection = await Collection.find_one(Collection.id == id, **LINK_POLICY)

    if collection is None:
        raise CollectionNotFound
    assert check_user_access(
        groups, collection.readers + collection.writers, scopes=scopes
    )
    if collection.products:
        collection_list = await collection_product_filter(groups, scopes, [collection])
        collection = collection_list[0]
    return collection


async def read_most_recent(
    groups: list[str],
    scopes: set[str],
    maximum: int = 16,
) -> list[Collection]:
    # TODO: Implement updated time for collections.
    if "hippo:admin" in scopes:
        access_query = {}
    else:
        access_query = {
            "$or": [{"readers": {"$in": groups}}, {"writers": {"$in": groups}}]
        }

    collection_list = await Collection.find(access_query, **LINK_POLICY).to_list(
        maximum
    )

    filtered_collection_list = await collection_product_filter(
        groups, scopes, collection_list
    )

    return filtered_collection_list


async def read_pinned(
    groups: list[str],
    scopes: set[str],
    maximum: int = 16,
) -> list[Collection]:
    if "hippo:admin" in scopes:
        access_query = {"pinned": True}
    else:
        access_query = {
            "$and": [
                {"pinned": True},
                {"$or": [{"readers": {"$in": groups}}, {"writers": {"$in": groups}}]},
            ]
        }

    collection_list = await Collection.find(access_query, **LINK_POLICY).to_list(
        maximum
    )

    return collection_list


async def search_by_name(
    name: str, groups: list[str], scopes: set[str]
) -> list[Collection]:
    """
    Search for Collections by name using the text index.
    """

    if "hippo:admin" in scopes:
        access_query = {}
    else:
        access_query = {
            "$or": [{"readers": {"$in": groups}}, {"writers": {"$in": groups}}]
        }

    results = (
        await Collection.find(access_query, Text(name), **LINK_POLICY)  # noqa: E712
        .sort([("score", {"$meta": "textScore"})])
        .to_list()
    )

    filtered_collection_list = await collection_product_filter(groups, scopes, results)

    return filtered_collection_list


async def search_by_owner(
    owner: str, groups: list[str], scopes: set[str]
) -> list[Collection]:
    """
    Search for Collections by owner; owner search is case insensitive
    but must otherwise be an exact match
    """

    owner_regex = {"$regex": f"^{re.escape(owner)}$", "$options": "i"}

    if "hippo:admin" in scopes:
        access_query = {}
    else:
        access_query = {
            "$or": [{"readers": {"$in": groups}}, {"writers": {"$in": groups}}]
        }

    results = await Collection.find(
        {**access_query, "owner": owner_regex}, **LINK_POLICY
    ).to_list()

    filtered_collection_list = await collection_product_filter(groups, scopes, results)

    return filtered_collection_list


async def diff(
    id: PydanticObjectId,
    access_groups: list[str],
    scopes: set[str],
    name: str | None = None,
    description: str | None = None,
):
    """
    Diff a collection against the current version.
    """
    collection = await read(id=id, groups=access_groups, scopes=scopes)

    diff_dict = {}

    if name and name != collection.name:
        diff_dict["name"] = [collection.name, name]

    if description and description != collection.description:
        diff_dict["description"] = [collection.description, description]

    return diff_dict


async def update(
    id: PydanticObjectId,
    access_groups: list[str],
    scopes: set[str],
    name: str | None,
    description: str | None,
):
    collection = await read(id=id, groups=access_groups, scopes=scopes)

    assert check_user_access(access_groups, collection.writers, scopes=scopes)

    if name:
        collection.name = name

    if description:
        collection.description = description

    await collection.save()

    return collection


async def add_child(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    groups: list[str],
    scopes: set[str],
) -> Collection:
    parent = await read(id=parent_id, groups=groups, scopes=scopes)
    child = await read(id=child_id, groups=groups, scopes=scopes)
    assert check_user_access(groups, parent.writers, scopes=scopes)
    parent.child_collections.append(child)
    await parent.save()

    return parent


async def remove_child(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    groups: list[str],
    scopes: set[str],
) -> Collection:
    parent = await read(id=parent_id, groups=groups, scopes=scopes)
    assert check_user_access(groups, parent.writers, scopes=scopes)
    parent.child_collections = [x for x in parent.child_collections if x.id != child_id]
    await parent.save()

    return parent


async def pin(
    id: PydanticObjectId,
    groups: list[str],
    scopes: set[str],
) -> Collection:
    collection = await read(id=id, groups=groups, scopes=scopes)
    assert check_user_access(groups, collection.writers, scopes=scopes)
    collection.pinned = True
    await collection.save()

    return collection


async def unpin(
    id: PydanticObjectId,
    groups: list[str],
    scopes: set[str],
) -> Collection:
    collection = await read(id=id, groups=groups, scopes=scopes)
    assert check_user_access(groups, collection.writers, scopes=scopes)
    collection.pinned = False
    await collection.save()

    return collection


async def delete(id: PydanticObjectId, groups: list[str], scopes: set[str]):
    collection = await read(id=id, groups=groups, scopes=scopes)
    assert check_user_access(
        groups, collection.readers + collection.writers, scopes=scopes
    )
    await collection.delete()

    return


async def collection_product_filter(
    groups: list[str],
    scopes: set[str],
    collection_list: list[Collection],
) -> list[Collection]:
    filtered_collection_list = []
    for collection in collection_list:
        if collection.products:
            product_list = []
            for product in collection.products:
                try:
                    assert check_user_access(
                        groups, product.readers + product.writers, scopes=scopes
                    )
                    product_list.append(product)
                except HTTPException:
                    continue
            collection.products = product_list
        filtered_collection_list.append(collection)
    return filtered_collection_list
